import datetime
import os
import pathlib

from io import BytesIO
from zoneinfo import ZoneInfo

import pandas as pd
import pytest

from panel import Param, bind
from panel.chat.icon import ChatReactionIcons
from panel.chat.message import ChatMessage, _FileInputMessage
from panel.layout import Column, Row
from panel.pane.image import PNG, SVG, Image
from panel.pane.markup import HTML, DataFrame, Markdown
from panel.pane.media import Audio
from panel.tests.util import mpl_available, mpl_figure
from panel.widgets.button import Button
from panel.widgets.input import (
    FileInput, IntInput, TextAreaInput, TextInput,
)

ASSETS = pathlib.Path(__file__).parent.parent / 'pane' / 'assets'

PNG_FILE = 'https://assets.holoviz.org/panel/samples/png_sample.png'
SVG_FILE = 'https://assets.holoviz.org/panel/samples/svg_sample.svg'

class TestChatMessage:
    def test_layout(self):
        message = ChatMessage(object="ABC", header_objects=["Header Test", "Header 2"], footer_objects=["Footer Test", "Footer 2"])
        columns = message._composite.objects
        assert len(columns) == 2

        avatar_pane = columns[0][0]
        assert isinstance(avatar_pane, HTML)
        assert avatar_pane.object == "🧑"

        meta_row = columns[1][0]
        user_pane = meta_row[0]
        assert isinstance(user_pane, HTML)
        assert user_pane.object == "User"

        header_row = columns[1][1]
        assert isinstance(header_row[0], Markdown)
        assert header_row[0].object == "Header Test"
        assert isinstance(header_row[1], Markdown)
        assert header_row[1].object == "Header 2"

        center_row = columns[1][2]
        assert isinstance(center_row, Row)

        object_pane = center_row[0]
        assert isinstance(object_pane, Markdown)
        assert object_pane.object == "ABC"

        icons = columns[1][4][1]
        assert isinstance(icons, ChatReactionIcons)

        footer_col = columns[1][3]
        assert isinstance(footer_col, Column)

        assert isinstance(footer_col[0], Markdown)
        assert footer_col[0].object == "Footer Test"
        assert isinstance(footer_col[1], Markdown)
        assert footer_col[1].object == "Footer 2"

        timestamp_pane = columns[1][5][0]
        assert isinstance(timestamp_pane, HTML)

    def test_reactions_dynamic(self):
        message = ChatMessage("hi", reactions=["favorite"])
        assert message.reaction_icons.value == ["favorite"]

        message.reactions = ["thumbs-up"]
        assert message.reaction_icons.value == ["thumbs-up"]

    def test_reaction_icons_dynamic(self):
        message = ChatMessage("hi", reaction_icons={"favorite": "heart"})
        assert message.reaction_icons.options == {"favorite": "heart"}

        message.reaction_icons = ChatReactionIcons(options={"like": "thumb-up"})
        assert message._icons_row[-1] == message.reaction_icons

        message.reaction_icons = ChatReactionIcons(options={})

        message = ChatMessage("hi", reaction_icons={})

    def test_reactions_link(self):
        # on init
        message = ChatMessage(reactions=["favorite"])
        assert message.reaction_icons.value == ["favorite"]

        # on change in message
        message.reactions = []
        assert message.reaction_icons.value == []

        # on change in reaction_icons
        message.reactions = ["favorite"]
        assert message.reaction_icons.value == ["favorite"]

    def test_reaction_icons_input_dict(self):
        message = ChatMessage(reaction_icons={"favorite": "heart"})
        assert isinstance(message.reaction_icons, ChatReactionIcons)
        assert message.reaction_icons.options == {"favorite": "heart"}

    def test_update_avatar(self):
        message = ChatMessage(avatar="A")
        columns = message._composite.objects
        avatar_pane = columns[0][0]
        assert isinstance(avatar_pane, HTML)
        assert avatar_pane.object == "A"

        message.avatar = "B"
        avatar_pane = columns[0][0]
        assert avatar_pane.object == "B"

        message.avatar = "❤️"
        avatar_pane = columns[0][0]
        assert avatar_pane.object == "❤️"

        message.avatar = "https://assets.holoviz.org/panel/samples/jpg_sample.jpg"
        avatar_pane = columns[0][0]
        assert isinstance(avatar_pane, Image)
        assert (
            avatar_pane.object
            == "https://assets.holoviz.org/panel/samples/jpg_sample.jpg"
        )

        message.show_avatar = False
        avatar_layout = columns[0]
        assert not avatar_layout.visible

        message.avatar = SVG(
            "https://tabler-icons.io/static/tabler-icons/icons/user.svg"
        )
        avatar_pane = columns[0][0]
        assert isinstance(avatar_pane, SVG)

    def test_update_user(self):
        message = ChatMessage(user="Andrew")
        columns = message._composite.objects
        user_pane = columns[1][0][0]
        assert isinstance(user_pane, HTML)
        assert user_pane.object == "Andrew"

        message.user = "August"
        user_pane = columns[1][0][0]
        assert user_pane.object == "August"

        message.show_user = False
        user_pane = columns[1][0][0]
        assert not user_pane.visible

    def test_update_object(self):
        message = ChatMessage(object="Test")
        columns = message._composite.objects
        object_pane = columns[1][2][0]
        assert isinstance(object_pane, Markdown)
        assert object_pane.object == "Test"

        message.object = TextInput(value="Also testing...")
        object_pane = columns[1][2][0]
        assert isinstance(object_pane, TextInput)
        assert object_pane.value == "Also testing..."

        message.object = _FileInputMessage(
            contents=b"I am a file", file_name="test.txt", mime_type="text/plain"
        )
        object_pane = columns[1][2][0]
        assert isinstance(object_pane, Markdown)
        assert object_pane.object == "I am a file"

    @pytest.mark.flaky(reruns=3, reason="Minute can change during test run")
    def test_update_timestamp(self):
        message = ChatMessage()
        columns = message._composite.objects
        timestamp_pane = columns[1][5][0]
        assert isinstance(timestamp_pane, HTML)
        dt_str = datetime.datetime.now().strftime("%H:%M")
        assert timestamp_pane.object == dt_str

        message = ChatMessage(timestamp_tz="UTC")
        columns = message._composite.objects
        timestamp_pane = columns[1][5][0]
        assert isinstance(timestamp_pane, HTML)
        dt_str = datetime.datetime.now(datetime.timezone.utc).strftime("%H:%M")
        assert timestamp_pane.object == dt_str

        message = ChatMessage(timestamp_tz="US/Pacific")
        columns = message._composite.objects
        timestamp_pane = columns[1][5][0]
        assert isinstance(timestamp_pane, HTML)
        dt_str = datetime.datetime.now(tz=ZoneInfo("US/Pacific")).strftime("%H:%M")
        assert timestamp_pane.object == dt_str

        special_dt = datetime.datetime(2023, 6, 24, 15)
        message.timestamp = special_dt
        timestamp_pane = columns[1][5][0]
        dt_str = special_dt.strftime("%H:%M")
        assert timestamp_pane.object == dt_str

        mm_dd_yyyy = "%b %d, %Y"
        message.timestamp_format = mm_dd_yyyy
        timestamp_pane = columns[1][5][0]
        dt_str = special_dt.strftime(mm_dd_yyyy)
        assert timestamp_pane.object == dt_str

        message.show_timestamp = False
        timestamp_pane = columns[1][5][0]
        assert not timestamp_pane.visible

    def test_does_not_turn_widget_into_str(self):
        button = Button()
        message = ChatMessage(object=button)
        assert message.object == button

    def test_include_stylesheets_inplace_on_layouts(self):
        message = ChatMessage(
            Row(Markdown("Hello", css_classes=["message"]), stylesheets=["row.css"]),
            stylesheets=["chat.css"]
        )
        assert message.stylesheets == ["chat.css"]
        assert message.object.stylesheets == message._stylesheets + ["chat.css", "row.css"]

        # # nested
        message = ChatMessage(
            Row(
                Row(Markdown("Hello", css_classes=["message"]), stylesheets=["row2.css"]),
                stylesheets=["row.css"]
            ),
            stylesheets=["chat.css"]
        )
        assert message.object.stylesheets == ChatMessage._stylesheets + ["chat.css", "row.css"]
        assert message.object.objects[0].stylesheets == ChatMessage._stylesheets + ["chat.css", "row2.css"]

    def test_include_message_css_class_inplace(self):
        # markdown
        message = ChatMessage(object=Markdown("hello"))
        assert message.object.css_classes == ["message"]

        # custom css class; no message appended
        message = ChatMessage(object=Markdown("hello", css_classes=["custom"]))
        assert message.object.css_classes == ["custom"]

        # nested in layout; message appended
        message = ChatMessage(object=Row(Markdown("hello")))
        assert message.object.objects[0].css_classes == ["message"]

        # nested in layout as a string; message appended
        message = ChatMessage(object=Row("hello"))
        assert message.object.objects[0].css_classes == ["message"]

        # nested in layout with custom css; no message appended
        message = ChatMessage(object=Row(Markdown("hello", css_classes=["custom"])))
        assert message.object.objects[0].css_classes == ["custom"]

    @mpl_available
    def test_can_display_any_python_object_that_panel_can_display(self):
        # For example matplotlib figures
        ChatMessage(object=mpl_figure())

        # For example async functions
        async def async_func():
            return "hello"

        ChatMessage(object=async_func)

        # For example async generators
        async def async_generator():
            yield "hello"
            yield "world"

        ChatMessage(object=async_generator)

    def test_can_use_pn_param_without_raising_exceptions(self):
        message = ChatMessage()
        Param(message)

    def test_bind_reactions(self):
        def callback(reactions):
            message.object = " ".join(reactions)

        message = ChatMessage(object="Hello")
        bind(callback, message.param.reactions, watch=True)
        message.reactions = ["favorite"]
        assert message.object == "favorite"

    def test_show_reaction_icons(self):
        message = ChatMessage()
        assert message.reaction_icons.visible
        message.show_reaction_icons = False
        assert not message.reaction_icons.visible

    def test_default_avatars(self):
        assert isinstance(ChatMessage.default_avatars, dict)
        assert (
            ChatMessage(user="Assistant").avatar == ChatMessage(user="assistant").avatar
        )
        assert ChatMessage(object="Hello", user="NoDefaultUserAvatar").avatar == ""

    def test_default_avatars_depends_on_user(self):
        ChatMessage.default_avatars["test1"] = "1"
        ChatMessage.default_avatars["test2"] = "2"

        message = ChatMessage(object="Hello", user="test1")
        assert message.avatar == "1"

        message.user = "test2"
        assert message.avatar == "2"

    def test_default_avatars_can_be_updated_but_the_original_stays(self):
        assert ChatMessage(user="Assistant").avatar == "🤖"
        ChatMessage.default_avatars["assistant"] = "👨"
        assert ChatMessage(user="Assistant").avatar == "👨"

        assert ChatMessage(user="System").avatar == "⚙️"

    def test_chat_copy_icon(self):
        message = ChatMessage(object="testing")
        assert message.chat_copy_icon.visible
        assert message.chat_copy_icon.value == "testing"

    @pytest.mark.parametrize("widget", [TextInput, TextAreaInput])
    def test_chat_copy_icon_text_widget(self, widget):
        message = ChatMessage(object=widget(value="testing"))
        assert message.chat_copy_icon.visible
        assert message.chat_copy_icon.value == "testing"

    def test_chat_copy_icon_disabled(self):
        message = ChatMessage(object="testing", show_copy_icon=False)
        assert not message.chat_copy_icon.visible
        assert not message.chat_copy_icon.value

    @pytest.mark.parametrize("component", [Column, FileInput])
    def test_chat_copy_icon_not_string(self, component):
        message = ChatMessage(object=component())
        assert not message.chat_copy_icon.visible
        assert not message.chat_copy_icon.value

    def test_serialize_text_prefix_with_viewable_type(self):
        message = ChatMessage(Markdown("string"))
        assert message.serialize() == "Markdown='string'"

    def test_serialize_text_prefix_with_viewable_name(self):
        message = ChatMessage(Markdown("string", name="Important Name"))
        assert message.serialize() == "Important Name='string'"

    def test_serialize_text_prefix_with_viewable_and_container_type(self):
        message = ChatMessage(Column(Markdown("string")))
        assert message.serialize() == "Column(Markdown='string')"

    def test_serialize_text_prefix_with_viewable_and_container_name(self):
        message = ChatMessage(
            Column(Markdown("string", name="Important Name"), name="Test Col")
        )
        assert message.serialize() == "Test Col(Important Name='string')"

    def test_serialize_multiple_objects_in_container(self):
        message = ChatMessage(
            Column(Markdown("markdown"), HTML("html"))
        )
        assert message.serialize() == "Column(\n    Markdown='markdown',\n    HTML='html'\n)"

    def test_serialize_nested_objects_in_container(self):
        message = ChatMessage(
            Row(
                Column(IntInput(name='int val')),
                Column(Markdown("markdown"), HTML("html"))
            )
        )
        assert message.serialize() == "Row(\n    Column(int val=0),\n    Column(\n        Markdown='markdown',\n        HTML='html'\n    )\n)"  # noqa

    def test_serialize_png_url(self):
        message = ChatMessage(PNG(PNG_FILE))
        assert message.serialize() == "PNG='https://assets.holoviz.org/panel/samples/png_sample.png'"

    @pytest.mark.internet
    def test_serialize_svg_embed(self):
        svg = SVG(SVG_FILE, embed=True, alt_text="abc")
        with BytesIO(svg._data(SVG_FILE)) as buf:
            message = ChatMessage(SVG(buf))
        assert message.serialize() == "SVG"

    @pytest.mark.skipif(os.name == 'nt', reason="Wrong Windows path")
    def test_serialize_audio(self):
        message = ChatMessage(Audio(str(ASSETS / 'mp3.mp3')))
        assert message.serialize() == f"Audio='{ASSETS / 'mp3.mp3'}'"

    def test_serialize_dataframe(self):
        message = ChatMessage(DataFrame(pd.DataFrame({'a': [1, 2, 3]})))
        assert message.serialize() == "DataFrame=   a\n0  1\n1  2\n2  3"



import numpy as np
import pandas as pd
import param

from panel.custom import PyComponent, ReactiveESM
from panel.layout import Row
from panel.pane import Markdown
from panel.viewable import Viewable


class SimplePyComponent(PyComponent):

    def __panel__(self):
        return Row(1, 2, 3, height=42)


def test_py_component_syncs(document, comm):
    spy = SimplePyComponent(width=42)

    spy.get_root(document, comm)

    assert isinstance(spy._view__, Row)
    assert spy._view__.width == 42
    assert spy.height == 42

    spy.width = 84

    assert spy._view__.width == 84

    spy._view__.width = 42

    assert spy._view__.width == 42


def test_py_component_cleanup(document, comm):
    spy = SimplePyComponent(width=42)

    model = spy.get_root(document, comm)

    assert model.ref['id'] in spy._models
    assert model.ref['id'] in spy._view__._models

    spy._cleanup(model)

    assert not spy._models
    assert not spy._view__._models


class ESMDataFrame(ReactiveESM):

    df = param.DataFrame()


def test_reactive_esm_sync_dataframe(document, comm):
    esm_df = ESMDataFrame()

    model = esm_df.get_root(document, comm)

    esm_df.df = pd.DataFrame({"1": [2]})

    assert isinstance(model.data.df, dict)
    assert len(model.data.df) == 2
    expected = {"index": np.array([0]), "1": np.array([2])}
    for col, values in model.data.df.items():
        np.testing.assert_array_equal(values, expected.get(col))


class ESMWithChildren(ReactiveESM):

    child = param.ClassSelector(class_=Viewable)

    children = param.List(item_type=Viewable)


def test_reactive_esm_model_cleanup(document, comm):
    esm = ReactiveESM()

    model = esm.get_root(document, comm)

    ref = model.ref['id']
    assert ref in esm._models
    assert esm._models[ref] == (model, None)

    esm._cleanup(model)
    assert esm._models == {}

def test_reactive_esm_child_model_cleanup(document, comm):
    md = Markdown('foo')
    esm = ESMWithChildren(child=md)

    model = esm.get_root(document, comm)

    ref = model.ref['id']
    assert ref in md._models

    md._cleanup(model)
    assert md._models == {}

def test_reactive_esm_child_model_cleanup_on_replace(document, comm):
    md1 = Markdown('foo')
    esm = ESMWithChildren(child=md1)

    model = esm.get_root(document, comm)

    ref = model.ref['id']
    assert ref in md1._models
    md1_model, _ = md1._models[ref]
    assert model.data.child is md1_model

    esm.child = md2 = Markdown('bar')

    assert md1._models == {}
    assert ref in md2._models
    md2_model, _ = md2._models[ref]
    assert model.data.child is md2_model

def test_reactive_esm_children_models_cleanup(document, comm):
    md = Markdown('foo')
    esm = ESMWithChildren(children=[md])

    model = esm.get_root(document, comm)

    ref = model.ref['id']
    assert ref in md._models

    md._cleanup(model)
    assert md._models == {}

def test_reactive_esm_children_models_cleanup_on_replace(document, comm):
    md1 = Markdown('foo')
    esm = ESMWithChildren(children=[md1])

    model = esm.get_root(document, comm)

    ref = model.ref['id']
    assert ref in md1._models
    md1_model, _ = md1._models[ref]
    assert model.data.children == [md1_model]

    esm.children = [md2] = [Markdown('bar')]

    assert md1._models == {}
    assert ref in md2._models
    md2_model, _ = md2._models[ref]
    assert model.data.children == [md2_model]

class ESMOverride(ReactiveESM):

    width = param.Integer(default=42)

def test_esm_parameter_override(document, comm):
    esm = ESMOverride()

    model = esm.get_root(document, comm)

    assert model.width == 42

    esm.width = 84

    assert model.width == 84



import param

import panel as pn

from panel.custom import ReactComponent

pn.extension()

class JSSlideshow(ReactComponent):

    index = param.Integer(default=0)

    _esm = """
	export function render({ model }) {
      const [index, setIndex] = model.useState("index")
      const img = `https://picsum.photos/800/300?image=${index}`
      return <img id="slideshow" src={img} onClick={ (event) => { setIndex(index+1) } }></img>
    }
	"""


class PySlideshow(ReactComponent):

    index = param.Integer(default=0)

    _esm = """
	export function render({ model }) {
      const [index, setIndex] = model.useState("index")
      const img = `https://picsum.photos/800/300?image=${index}`
      return <img id="slideshow" src={img} onClick={ (event) => model.send_event('click', event) }></img>
    }
	"""

    def _handle_click(self, event):
        self.index += 1


pn.Row(
    JSSlideshow(),
    PySlideshow()
).servable()



from bokeh.command.subcommand import Argument, Subcommand

from ..io.compile import RED, compile_components, find_module_bundles


class Compile(Subcommand):
    ''' Subcommand to generate a new encryption key.

    '''

    #: name for this subcommand
    name = "compile"

    help = "Compiles an ESM component using node and esbuild"

    args = (
        ('modules', Argument(
            metavar = 'DIRECTORY-OR-SCRIPT',
            nargs   = "*",
            help    = "The Python modules to compile. May optionally define a single class.",
            default = None,
        )),
        ('--build-dir', dict(
            action = 'store',
            type    = str,
            help    = "Where to write the build directory."
        )),
        ('--unminified', dict(
            action  = 'store_true',
            help    = "Whether to generate unminified output."
        )),
        ('--verbose', dict(
            action  = 'store_true',
            help    = "Whether to show verbose output. Note when setting --outfile only the result will be printed to stdout."
        )),
    )

    def invoke(self, args):
        bundles = {}
        for module in args.modules:
            try:
                module_bundles = find_module_bundles(module)
            except RuntimeError as e:
                print(f'{RED} {e}')  # noqa
                return 1
            if not module_bundles:
                print (  # noqa
                    f'{RED} Could not find any ESM components to compile '
                    f'in {module}, ensure you provided the right module.'
                )
            for bundle, components in module_bundles.items():
                if bundle in bundles:
                    bundles[bundle] += components
                else:
                    bundles[bundle] = components

        errors = 0
        for bundle, components in bundles.items():
            component_names = '\n- '.join(c.name for c in components)
            print(f"Building {bundle} containing the following components:\n\n- {component_names}\n")  # noqa
            out = compile_components(
                components,
                build_dir=args.build_dir,
                minify=not args.unminified,
                outfile=bundle,
                verbose=args.verbose,
            )
            if not out:
                errors += 1
        return int(errors>0)



import pathlib

from bokeh.models import ImportedStyleSheet

from panel.io.resources import CDN_DIST
from panel.theme.base import BOKEH_DARK, Design, Inherit
from panel.viewable import Viewable
from panel.widgets import FloatSlider, IntSlider, TextInput


def _custom_repr(self):
    try:
        return f"ImportedStyleSheet(url={self.url!r})"
    except Exception:
        return "ImportedStyleSheet(...)"

ImportedStyleSheet.__repr__ = _custom_repr


class DesignTest(Design):

    modifiers = {
        IntSlider: {
            'stylesheets': [Inherit, 'http://example.com/baz.css']
        },
        FloatSlider: {
            'styles': {'color': 'red'},
            'stylesheets': [Inherit, pathlib.Path(__file__).parent / 'test.css']
        },
        TextInput: {
            'styles': {'color': 'green'},
            'stylesheets': [Inherit, '../theme/test.css']
        },
        Viewable: {
            'stylesheets': ['foo.css']
        }
    }


def test_design_params():
    widget = TextInput()

    params, _ = DesignTest().params(widget)
    assert 'stylesheets' in params
    assert len(params['stylesheets']) == 3
    s1, s2, s3 = params['stylesheets']
    assert isinstance(s1, ImportedStyleSheet)
    assert s1.url == f'{CDN_DIST}bundled/theme/default.css'
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url == 'foo.css'
    assert s3 == '.bk-input {\n  color: red;\n}\n'

    assert params.get('styles') == {'color': 'green'}

def test_design_params_server(server_document):
    widget = TextInput()

    params, _ = DesignTest().params(widget)
    assert 'stylesheets' in params
    assert len(params['stylesheets']) == 3
    s1, s2, s3 = params['stylesheets']
    assert isinstance(s1, ImportedStyleSheet)
    assert s1.url == f'{CDN_DIST}bundled/theme/default.css'
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url == 'foo.css'
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url == 'components/test_base/DesignTest/modifiers/test.css'

    assert params.get('styles') == {'color': 'green'}

def test_design_params_inherited():
    widget = FloatSlider()

    params, _ = DesignTest().params(widget)
    assert 'stylesheets' in params
    assert len(params['stylesheets']) == 3
    s1, s2, s3 = params['stylesheets']
    assert isinstance(s1, ImportedStyleSheet)
    assert s1.url == f'{CDN_DIST}bundled/theme/default.css'
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url == 'foo.css'
    assert s3 == '.bk-input {\n  color: red;\n}\n'

    assert params.get('styles') == {'color': 'red'}

def test_design_params_inherited_server(server_document):
    widget = FloatSlider()

    params, _ = DesignTest().params(widget)
    assert 'stylesheets' in params
    assert len(params['stylesheets']) == 3
    s1, s2, s3 = params['stylesheets']
    assert isinstance(s1, ImportedStyleSheet)
    assert s1.url == f'{CDN_DIST}bundled/theme/default.css'
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url == 'foo.css'
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url == 'components/test_base/DesignTest/modifiers/test.css'

    assert params.get('styles') == {'color': 'red'}

def test_design_params_url_inherited():
    widget = IntSlider()

    params, _ = DesignTest().params(widget)
    assert 'stylesheets' in params
    assert len(params['stylesheets']) == 3
    s1, s2, s3 = params['stylesheets']
    assert isinstance(s1, ImportedStyleSheet)
    assert s1.url == f'{CDN_DIST}bundled/theme/default.css'
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url == 'foo.css'
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url == 'http://example.com/baz.css'

def test_design_apply(document, comm):
    widget = TextInput()
    model = widget.get_root(document, comm=comm)

    DesignTest().apply(widget, model)

    assert len(model.stylesheets) == 5
    s1, s2, s3, s4, s5 = model.stylesheets
    assert isinstance(s1, str)
    assert 'pn-loading' in s1
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url.endswith('/dist/css/loading.css')
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url.endswith('/dist/bundled/theme/default.css')
    assert isinstance(s4, ImportedStyleSheet)
    assert s4.url.endswith('foo.css')
    assert s5 == '.bk-input {\n  color: red;\n}\n'
    assert model.styles == {'color': 'green'}

def test_design_apply_not_isolated(document, comm):
    widget = TextInput()
    model = widget.get_root(document, comm=comm)

    DesignTest().apply(widget, model, isolated=False)

    assert len(model.stylesheets) == 4
    s1, s2, s3, s4 = model.stylesheets
    assert isinstance(s1, str)
    assert 'pn-loading' in s1
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url.endswith('/dist/css/loading.css')
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url.endswith('foo.css')
    assert s4 == '.bk-input {\n  color: red;\n}\n'

    assert model.styles == {'color': 'green'}

def test_design_apply_inherited(document, comm):
    widget = FloatSlider()
    model = widget.get_root(document, comm=comm)

    DesignTest().apply(widget, model)

    assert len(model.stylesheets) == 5
    s1, s2, s3, s4, s5 = model.stylesheets
    assert isinstance(s1, str)
    assert 'pn-loading' in s1
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url.endswith('/dist/css/loading.css')
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url.endswith('/dist/bundled/theme/default.css')
    assert isinstance(s4, ImportedStyleSheet)
    assert s4.url.endswith('foo.css')
    assert s5 == '.bk-input {\n  color: red;\n}\n'

    assert model.styles == {'color': 'red'}

def test_design_apply_url_inherited(document, comm):
    widget = IntSlider()
    model = widget.get_root(document, comm=comm)

    DesignTest().apply(widget, model)

    assert len(model.stylesheets) == 5
    s1, s2, s3, s4, s5 = model.stylesheets
    assert isinstance(s1, str)
    assert 'pn-loading' in s1
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url.endswith('/dist/css/loading.css')
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url.endswith('/dist/bundled/theme/default.css')
    assert isinstance(s4, ImportedStyleSheet)
    assert s4.url.endswith('foo.css')
    assert isinstance(s5, ImportedStyleSheet)
    assert s5.url == 'http://example.com/baz.css'

def test_design_apply_with_dark_theme(document, comm):
    widget = TextInput()
    model = widget.get_root(document, comm=comm)
    model.document = document

    DesignTest(theme='dark').apply(widget, model)

    assert len(model.stylesheets) == 5
    s1, s2, s3, s4, s5 = model.stylesheets
    assert isinstance(s1, str)
    assert 'pn-loading' in s1
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url.endswith('/dist/css/loading.css')
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url.endswith('/dist/bundled/theme/dark.css')
    assert isinstance(s4, ImportedStyleSheet)
    assert s4.url.endswith('foo.css')
    assert s5 == '.bk-input {\n  color: red;\n}\n'

    assert document.theme._json == BOKEH_DARK

def test_design_apply_with_dark_theme_not_isolated(document, comm):
    widget = TextInput()
    model = widget.get_root(document, comm=comm)
    model.document = document

    DesignTest(theme='dark').apply(widget, model, isolated=False)

    assert len(model.stylesheets) == 4
    s1, s2, s3, s4 = model.stylesheets
    assert isinstance(s1, str)
    assert 'pn-loading' in s1
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url.endswith('/dist/css/loading.css')
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url.endswith('foo.css')
    assert s4 == '.bk-input {\n  color: red;\n}\n'

    assert document.theme._json == BOKEH_DARK

def test_design_apply_with_dist_url(document, comm):
    widget = TextInput()
    model = widget.get_root(document, comm=comm)
    model.document = document
    document._template_variables['dist_url'] = 'https://mock.holoviz.org/'

    DesignTest().apply(widget, model)

    assert len(model.stylesheets) == 5
    s1, s2, s3, s4, s5 = model.stylesheets
    assert isinstance(s1, str)
    assert 'pn-loading' in s1
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url.startswith('https://mock.holoviz.org/css/loading.css?v')
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url.startswith('https://mock.holoviz.org/bundled/theme/default.css?v')
    assert isinstance(s4, ImportedStyleSheet)
    assert s4.url == 'foo.css'
    assert s5 == '.bk-input {\n  color: red;\n}\n'

    assert model.styles == {'color': 'green'}

def test_design_apply_shared_stylesheet_models(document, comm):
    widget1 = TextInput()
    model1 = widget1.get_root(document, comm=comm)
    model1.document = document
    widget2 = TextInput()
    model2 = widget2.get_root(document, comm=comm)
    model2.document = document

    DesignTest().apply(widget1, model1)
    DesignTest().apply(widget2, model2)

    assert widget1.stylesheets == widget2.stylesheets



from panel.custom import Child, JSComponent


class Example(JSComponent):

    child = Child()

    _esm = """
    export function render({ model }) {
      return model.get_child('child')
    }"""

Example(child='A Markdown pane!').servable()



import os
import re
import tempfile
import time

from textwrap import dedent

import pytest
import requests

from panel.tests.util import (
    NBSR, linux_only, run_panel_serve, unix_only, wait_for_port,
    wait_for_regex, write_file,
)


@linux_only
def test_autoreload_app(py_file):
    app = "import panel as pn; pn.Row('# Example').servable(title='A')"
    app2 = "import panel as pn; pn.Row('# Example 2').servable(title='B')"
    write_file(app, py_file.file)

    app_name = os.path.basename(py_file.name)[:-3]

    with run_panel_serve(["--port", "0", '--autoreload', py_file.name]) as p:
        port = wait_for_port(p.stdout)
        r = requests.get(f"http://localhost:{port}/{app_name}")
        assert r.status_code == 200
        assert "<title>A</title>" in r.content.decode('utf-8')

        write_file(app2, py_file.file)

        r2 = requests.get(f"http://localhost:{port}/{app_name}")
        assert r2.status_code == 200
        assert "<title>B</title>" in r2.content.decode('utf-8')


@linux_only
def test_autoreload_app_local_module(py_files):
    py_file1, py_file2 = py_files
    app_name = os.path.basename(py_file1.name)[:-3]
    mod_name = os.path.basename(py_file2.name)[:-3]
    app = f"import panel as pn; from {mod_name} import title; pn.Row('# Example').servable(title=title)"
    write_file(app, py_file1.file)
    write_file("title = 'A'", py_file2.file)

    with run_panel_serve(["--port", "0", '--autoreload', py_file1.name]) as p:
        port = wait_for_port(p.stdout)
        r = requests.get(f"http://localhost:{port}/{app_name}")
        assert r.status_code == 200
        assert "<title>A</title>" in r.content.decode('utf-8')

        write_file("title = 'B'", py_file2.file)
        py_file2.file.close()
        time.sleep(1)

        r2 = requests.get(f"http://localhost:{port}/{app_name}")
        assert r2.status_code == 200
        assert "<title>B</title>" in r2.content.decode('utf-8')


@linux_only
def test_serve_admin(py_file):
    app = "import panel as pn; pn.Row('# Example').servable(title='A')"
    write_file(app, py_file.file)

    with run_panel_serve(["--port", "0", '--admin', py_file.name]) as p:
        port = wait_for_port(p.stdout)
        r = requests.get(f"http://localhost:{port}/admin")
        assert r.status_code == 200
        assert "Admin" in r.content.decode('utf-8')

@linux_only
def test_serve_admin_custom_endpoint(py_file):
    app = "import panel as pn; pn.Row('# Example').servable(title='A')"
    write_file(app, py_file.file)

    with run_panel_serve(["--port", "0", '--admin', '--admin-endpoint', 'foo', py_file.name]) as p:
        port = wait_for_port(p.stdout)
        r = requests.get(f"http://localhost:{port}/foo")
        assert r.status_code == 200
        assert "Admin" in r.content.decode('utf-8')
        r2 = requests.get(f"http://localhost:{port}/admin")
        assert r2.status_code == 404

@linux_only
@pytest.mark.parametrize('relative', [True, False])
def test_custom_html_index(relative, html_file):
    index = '<html><body>Foo</body></html>'
    write_file(index, html_file.file)
    app = "import panel as pn; pn.Row('# Example').servable(title='A')"
    app2 = "import panel as pn; pn.Row('# Example 2').servable(title='B')"
    py1 = tempfile.NamedTemporaryFile(mode='w', suffix='.py')
    py2 = tempfile.NamedTemporaryFile(mode='w', suffix='.py')
    write_file(app, py1.file)
    write_file(app2, py2.file)

    if relative:
        index_path = os.path.basename(html_file.name)
        cwd = os.path.dirname(html_file.name)
    else:
        index_path = html_file.name
        cwd = None

    with run_panel_serve(["--port", "0", "--index", index_path, py1.name, py2.name], cwd=cwd) as p:
        port = wait_for_port(p.stdout)
        r = requests.get(f"http://localhost:{port}/")
        assert r.status_code == 200
        assert r.content.decode('utf-8') == index

md_app = """
# My app

```python
import panel as pn
pn.extension(template='fast')
```

A description

```python
pn.Row('# Example').servable()
```
"""

@linux_only
def test_serve_markdown():
    md = tempfile.NamedTemporaryFile(mode='w', suffix='.md')
    write_file(md_app, md.file)

    with run_panel_serve(["--port", "0", md.name]) as p:
        port = wait_for_port(p.stdout)
        r = requests.get(f"http://localhost:{port}/")
        assert r.status_code == 200
        assert '<title>My app</title>' in r.content.decode('utf-8')


@unix_only
@pytest.mark.parametrize("arg", ["--warm", "--autoreload"])
def test_serve_num_procs(arg, tmp_path):
    app = "import panel as pn; pn.panel('Hello').servable()"
    py = tmp_path / "app.py"
    py.write_text(app)

    regex = re.compile(r'Starting Bokeh server with process id: (\d+)')
    with run_panel_serve(["--port", "0", py, "--num-procs", 2, arg], cwd=tmp_path) as p:
        pid1, pid2 = wait_for_regex(p.stdout, regex=regex, count=2)
        assert pid1 != pid2


@unix_only
def test_serve_num_procs_setup(tmp_path):
    app = "import panel as pn; pn.panel('Hello').servable()"
    py = tmp_path / "app.py"
    py.write_text(app)

    setup_app = 'import os; print(f"Setup PID {os.getpid()}", flush=True)'
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(setup_app)

    regex = re.compile(r'Setup PID (\d+)')
    with run_panel_serve(["--port", "0", py, "--num-procs", 2, "--setup", setup_py], cwd=tmp_path) as p:
        pid1, pid2 = wait_for_regex(p.stdout, regex=regex, count=2)
        assert pid1 != pid2


def test_serve_setup(tmp_path):
    app = "import panel as pn; pn.panel('Hello').servable()"
    py = tmp_path / "app.py"
    py.write_text(app)

    setup_app = 'print(f"Setup running before", flush=True)'
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(setup_app)

    regex = re.compile('(Setup running before)')
    with run_panel_serve(["--port", "0", py, "--setup", setup_py], cwd=tmp_path) as p:
        _, output = wait_for_regex(p.stdout, regex=regex, return_output=True)
        assert output[0].strip() == "Setup running before"


def test_serve_authorize_callback_exception(tmp_path):
    app = "import panel as pn; pn.panel('Hello').servable()"
    py = tmp_path / "app.py"
    py.write_text(app)

    setup_app = """\
        import panel as pn
        def auth(userinfo):
            raise ValueError("This is an error")
        pn.config.authorize_callback = auth"""
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(dedent(setup_app))

    regex = re.compile('(Authorization callback errored)')
    with run_panel_serve(["--port", "0", py, "--setup", setup_py], cwd=tmp_path) as p:
        nsbr = NBSR(p.stdout)
        port = wait_for_port(nsbr)
        resp = requests.get(f"http://localhost:{port}/")
        wait_for_regex(nsbr, regex=regex)
        assert resp.status_code == 403



"""
The message module provides a low-level API for rendering chat messages.
"""

from __future__ import annotations

import datetime

from contextlib import ExitStack
from dataclasses import dataclass
from functools import partial
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import (
    TYPE_CHECKING, Any, Callable, ClassVar, Optional, Union,
)
from zoneinfo import ZoneInfo

import param

from ..io.resources import CDN_DIST, get_dist_path
from ..io.state import state
from ..layout import Column, Row
from ..pane.base import Pane, ReplacementPane, panel as _panel
from ..pane.image import (
    PDF, FileBase, Image, ImageBase,
)
from ..pane.markup import (
    HTML, DataFrame, HTMLBasePane, Markdown,
)
from ..pane.media import Audio, Video
from ..param import ParamFunction
from ..viewable import Viewable
from ..widgets.base import Widget
from .icon import ChatCopyIcon, ChatReactionIcons
from .utils import (
    avatar_lookup, build_avatar_pane, serialize_recursively, stream_to,
)

if TYPE_CHECKING:
    from bokeh.document import Document
    from bokeh.model import Model
    from pyviz_comms import Comm

Avatar = Union[str, BytesIO, bytes, ImageBase]
AvatarDict = dict[str, Avatar]

USER_LOGO = "🧑"
ASSISTANT_LOGO = "🤖"
SYSTEM_LOGO = "⚙️"
ERROR_LOGO = "❌"
HELP_LOGO = "❓"
INPUT_LOGO = "❗"
GPT_3_LOGO = "{dist_path}assets/logo/gpt-3.svg"
GPT_4_LOGO = "{dist_path}assets/logo/gpt-4.svg"
WOLFRAM_LOGO = "{dist_path}assets/logo/wolfram.svg"
LUMEN_LOGO = "{dist_path}assets/logo/lumen.svg"
HOLOVIEWS_LOGO = "{dist_path}assets/logo/holoviews.svg"
HVPLOT_LOGO = "{dist_path}assets/logo/hvplot.svg"
PANEL_LOGO = "{dist_path}images/icon-vector.svg"

DEFAULT_AVATARS = {
    # User
    "client": USER_LOGO,
    "customer": USER_LOGO,
    "employee": USER_LOGO,
    "human": USER_LOGO,
    "person": USER_LOGO,
    "user": USER_LOGO,
    # Assistant
    "agent": ASSISTANT_LOGO,
    "ai": ASSISTANT_LOGO,
    "assistant": ASSISTANT_LOGO,
    "bot": ASSISTANT_LOGO,
    "chatbot": ASSISTANT_LOGO,
    "machine": ASSISTANT_LOGO,
    "robot": ASSISTANT_LOGO,
    # System
    "system": SYSTEM_LOGO,
    "exception": ERROR_LOGO,
    "error": ERROR_LOGO,
    "help": HELP_LOGO,
    "input": INPUT_LOGO,
    # Human
    "adult": "🧑",
    "baby": "👶",
    "boy": "👦",
    "child": "🧒",
    "girl": "👧",
    "man": "👨",
    "woman": "👩",
    # Machine
    "chatgpt": GPT_3_LOGO,
    "gpt3": GPT_3_LOGO,
    "gpt4": GPT_4_LOGO,
    "dalle": GPT_4_LOGO,
    "openai": GPT_4_LOGO,
    "huggingface": "🤗",
    "calculator": "🧮",
    "langchain": "🦜",
    "retriever": "📄",
    "tool": "🛠️",
    "translator": "🌐",
    "wolfram": WOLFRAM_LOGO,
    "wolfram alpha": WOLFRAM_LOGO,
    # Llama
    "llama": "🦙",
    "llama2": "🐪",
    # Plotting
    "plot": "📊",
    "lumen": LUMEN_LOGO,
    "holoviews": HOLOVIEWS_LOGO,
    "hvplot": HVPLOT_LOGO,
    "panel": PANEL_LOGO,
}


@dataclass
class _FileInputMessage:
    """
    A dataclass to hold the contents of a file input message.

    Parameters
    ----------
    contents : bytes
        The contents of the file.
    file_name : str
        The name of the file.
    mime_type : str
        The mime type of the file.
    """

    contents: bytes
    file_name: str
    mime_type: str


class ChatMessage(Pane):
    """
    Renders another component as a chat message with an associated user
    and avatar with support for various content types.

    This widget provides a structured view of chat messages, including features like:

    - Displaying user avatars, which can be text, emoji, or images.
    - Showing the user's name.
    - Displaying the message timestamp in a customizable format.
    - Associating reactions with messages and mapping them to icons.
    - Rendering various content types including text, images, audio, video, and more.

    Reference: https://panel.holoviz.org/reference/chat/ChatMessage.html

    :Example:

    >>> ChatMessage(object="Hello world!", user="New User", avatar="😊")
    """

    avatar = param.ClassSelector(default="", class_=(str, BytesIO, bytes, ImageBase), doc="""
        The avatar to use for the user. Can be a single character
        text, an emoji, or anything supported by `pn.pane.Image`. If
        not set, checks if the user is available in the
        default_avatars mapping; else uses the first character of the
        name.""")

    avatar_lookup = param.Callable(default=None, doc="""
        A function that can lookup an `avatar` from a user name. The
        function signature should be `(user: str) -> Avatar`. If this
        is set, `default_avatars` is disregarded.""")

    css_classes = param.List(default=["chat-message"],doc="""
        The CSS classes to apply to the widget.""")

    default_avatars = param.Dict(default=DEFAULT_AVATARS, doc="""
        A default mapping of user names to their corresponding avatars
        to use when the user is specified but the avatar is. You can
        modify, but not replace the dictionary.""")

    footer_objects = param.List(doc="""
        A list of objects to display in the column of the footer of the message.""")

    header_objects = param.List(doc="""
        A list of objects to display in the row of the header of the message.""")

    max_width = param.Integer(default=1200, bounds=(0, None), allow_None=True)

    object = param.Parameter(allow_refs=False, doc="""
        The message contents. Can be any Python object that panel can display.""")

    reactions = param.List(doc="""
        Reactions to associate with the message.""")

    reaction_icons = param.ClassSelector(class_=ChatReactionIcons, doc="""
        A mapping of reactions to their reaction icons; if not provided
        defaults to `{"favorite": "heart"}`.""", allow_refs=False)

    timestamp = param.Date(doc="""
        Timestamp of the message. Defaults to the creation time.""")

    timestamp_format = param.String(default="%H:%M", doc="The timestamp format.")

    timestamp_tz = param.String(default=None, doc="""
        The timezone to used for the creation timestamp; only applicable
        if timestamp is not set. If None, tries to use pn.state.browser_info.timezone,
        else, the default tz of datetime.datetime.now(); see `zoneinfo.available_timezones()`
        for a list of valid timezones.
    """)

    show_avatar = param.Boolean(default=True, doc="""
         Whether to display the avatar of the user.""")

    show_user = param.Boolean(default=True, doc="""
        Whether to display the name of the user.""")

    show_timestamp = param.Boolean(default=True, doc="""
        Whether to display the timestamp of the message.""")

    show_reaction_icons = param.Boolean(default=True, doc="""
        Whether to display the reaction icons.""")

    show_copy_icon = param.Boolean(default=True, doc="""
        Whether to display the copy icon.""")

    show_activity_dot = param.Boolean(default=False, doc="""
        Whether to show the activity dot.""")

    renderers = param.HookList(doc="""
        A callable or list of callables that accept the object and return a
        Panel object to render the object. If a list is provided, will
        attempt to use the first renderer that does not raise an
        exception. If None, will attempt to infer the renderer
        from the object."""
    )

    user = param.Parameter(default="User", doc="""
        Name of the user who sent the message.""")

    _stylesheets: ClassVar[list[str]] = [f"{CDN_DIST}css/chat_message.css"]

    # Declares whether Pane supports updates to the Bokeh model
    _updates: ClassVar[bool] = True

    def __init__(self, object=None, **params):
        self._exit_stack = ExitStack()
        self.chat_copy_icon = ChatCopyIcon(
            visible=False, width=15, height=15, css_classes=["copy-icon"]
        )
        if params.get("timestamp") is None:
            tz = params.get("timestamp_tz")
            if tz is not None:
                tz = ZoneInfo(tz)
            elif state.browser_info and state.browser_info.timezone:
                tz = ZoneInfo(state.browser_info.timezone)
            params["timestamp"] = datetime.datetime.now(tz=tz)

        reaction_icons = params.get("reaction_icons", {"favorite": "heart"})
        if isinstance(reaction_icons, dict):
            params["reaction_icons"] = ChatReactionIcons(options=reaction_icons, default_layout=Row, sizing_mode=None)
        self._internal = True
        super().__init__(object=object, **params)
        self.chat_copy_icon = ChatCopyIcon(
            visible=False, width=15, height=15, css_classes=["copy-icon"],
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
        )
        if not self.avatar:
            self._update_avatar()
        self._build_layout()

    def _build_layout(self):
        self._left_col = left_col = Column(
            self._render_avatar(),
            max_width=60,
            height=100,
            css_classes=["left"],
            visible=self.param.show_avatar,
            sizing_mode=None,
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
        )
        self.param.watch(self._update_avatar_pane, "avatar")

        self._object_panel = self._create_panel(self.object)
        self._update_chat_copy_icon()
        self._center_row = Row(
            self._object_panel,
            css_classes=["center"],
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
            sizing_mode=None
        )
        self.param.watch(self._update_object_pane, "object")
        self.param.watch(self._update_reaction_icons, "reaction_icons")

        self._user_html = HTML(
            self.param.user, height=20,
            css_classes=["name"],
            visible=self.param.show_user,
            sizing_mode=None,
        )

        self._activity_dot = HTML(
            "●",
            css_classes=["activity-dot"],
            margin=(5, 0),
            sizing_mode=None,
            visible=self.param.show_activity_dot,
        )

        meta_row = Row(
            self._user_html,
            self._activity_dot,
            css_classes=["meta"],
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
        )

        header_col = Column(
            objects=self.param.header_objects.rx(),
            sizing_mode="stretch_width",
            css_classes=["header"],
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
        )

        footer_col = Column(
            objects=self.param.footer_objects.rx(),
            sizing_mode="stretch_width",
            css_classes=["footer"],
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
        )

        self._timestamp_html = HTML(
            self.param.timestamp.rx().strftime(self.param.timestamp_format),
            css_classes=["timestamp"],
            visible=self.param.show_timestamp,
        )

        self._icons_row = Row(
            self.chat_copy_icon,
            self._render_reaction_icons(),
            css_classes=["icons"],
            sizing_mode="stretch_width",
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
        )

        self._right_col = right_col = Column(
            meta_row,
            header_col,
            self._center_row,
            footer_col,
            self._icons_row,
            self._timestamp_html,
            css_classes=["right"],
            sizing_mode=None,
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
        )
        viewable_params = {
            p: self.param[p] for p in self.param if p in Viewable.param
            if p in Viewable.param and p != 'name'
        }
        viewable_params['stylesheets'] = self._stylesheets + self.param.stylesheets.rx()
        self._composite = Row(left_col, right_col, **viewable_params)

    def __str__(self) -> str:
        """
        Serialize the message object to a string.
        """
        return str(self.serialize())

    @property
    def _synced_params(self) -> list[str]:
        return []

    def _get_model(
        self, doc: Document, root: Model | None = None,
        parent: Model | None = None, comm: Comm | None = None
    ) -> Model:
        model = self._composite._get_model(doc, root, parent, comm)
        if not comm:
            # If not in a notebook environment we potentially need to
            # replace path to avatar with server URL
            avatar = model.children[0].children[0]
            avatar.text = avatar.text.replace(CDN_DIST, get_dist_path())
        ref = (root or model).ref['id']
        self._models[ref] = (model, parent)
        return model

    def _select_renderer(
        self,
        contents: Any,
        mime_type: str,
    ):
        """
        Determine the renderer to use based on the mime type.
        """
        renderer = _panel
        if mime_type == "application/pdf":
            contents = self._exit_stack.enter_context(BytesIO(contents))
            renderer = partial(PDF, embed=True)
        elif mime_type.startswith("audio/"):
            file = self._exit_stack.enter_context(
                NamedTemporaryFile(suffix=".mp3", delete=False)
            )
            file.write(contents)
            file.seek(0)
            contents = file.name
            renderer = Audio
        elif mime_type.startswith("video/"):
            contents = self._exit_stack.enter_context(BytesIO(contents))
            renderer = Video
        elif mime_type.startswith("image/"):
            contents = self._exit_stack.enter_context(BytesIO(contents))
            renderer = Image
        elif mime_type.endswith("/csv"):
            import pandas as pd

            with BytesIO(contents) as buf:
                contents = pd.read_csv(buf)
            renderer = DataFrame
        elif mime_type.startswith("text"):
            if isinstance(contents, bytes):
                contents = contents.decode("utf-8")
        return contents, renderer

    def _include_stylesheets_inplace(self, obj):
        if hasattr(obj, "objects"):
            obj.objects[:] = [
                self._include_stylesheets_inplace(o) for o in obj.objects
            ]
        else:
            obj = _panel(obj)
        obj.stylesheets = [
            stylesheet for stylesheet in self._stylesheets + self.stylesheets
            if stylesheet not in obj.stylesheets
        ] + obj.stylesheets
        return obj

    def _include_message_css_class_inplace(self, obj):
        if hasattr(obj, "objects"):
            obj.objects[:] = [
                self._include_message_css_class_inplace(o)
                for o in obj.objects
            ]
        else:
            obj = _panel(obj)
        is_markup = isinstance(obj, HTMLBasePane) and not isinstance(obj, FileBase)
        if obj.css_classes or not is_markup:
            return obj
        if len(str(obj.object)) > 0:  # only show a background if there is content
            obj.css_classes = [*(css for css in obj.css_classes if css != "message"), "message"]
        obj.sizing_mode = None
        return obj

    def _set_params(self, obj, **params):
        """
        Set the sizing mode and height of the object.
        """
        self._include_stylesheets_inplace(obj)
        is_markup = isinstance(obj, HTMLBasePane) and not isinstance(obj, FileBase)
        if is_markup:
            self._include_message_css_class_inplace(obj)
        else:
            if obj.sizing_mode is None and not obj.width:
                params['sizing_mode'] = "stretch_width"

            if obj.height is None and not isinstance(obj, ParamFunction):
                params['height'] = 500
        obj.param.update(params)

    @staticmethod
    def _is_widget_renderer(renderer):
        return isinstance(renderer, type) and issubclass(renderer, Widget)

    def _create_panel(self, value, old=None):
        """
        Create a panel object from the value.
        """
        if isinstance(value, Viewable):
            self._internal = False
            self._include_stylesheets_inplace(value)
            self._include_message_css_class_inplace(value)
            return value

        renderer = None
        if isinstance(value, _FileInputMessage):
            contents = value.contents
            mime_type = value.mime_type
            value, renderer = self._select_renderer(contents, mime_type)
        else:
            try:
                import magic

                mime_type = magic.from_buffer(value, mime=True)
                value, renderer = self._select_renderer(value, mime_type)
            except Exception:
                pass

        renderers = self.renderers.copy() or []
        if renderer is not None:
            renderers.append(renderer)
        for renderer in renderers:
            try:
                if self._is_widget_renderer(renderer):
                    object_panel = renderer(value=value)
                else:
                    object_panel = renderer(value)
                if isinstance(object_panel, Viewable):
                    break
            except Exception:
                pass
        else:
            if isinstance(old, Markdown) and isinstance(value, str):
                self._set_params(old, enable_streaming=True, object=value)
                return old
            object_panel = _panel(value)

        self._set_params(object_panel)
        if type(old) is type(object_panel) and self._internal:
            ReplacementPane._recursive_update(old, object_panel)
            return object_panel

        self._internal = True
        return object_panel

    def _render_avatar(self) -> HTML | Image:
        """
        Render the avatar pane as some HTML text or Image pane.
        """
        avatar = self.avatar
        if not avatar and self.user:
            avatar = self.user[0]

        avatar_pane = build_avatar_pane(
            avatar, css_classes=["avatar"], height=35, width=35
        )
        return avatar_pane

    def _update_avatar_pane(self, event=None):
        new_avatar = self._render_avatar()
        old_type = type(self._left_col[0])
        new_type = type(new_avatar)
        if isinstance(event.old, (HTML, ImageBase)) or new_type is not old_type:
            self._left_col[:] = [new_avatar]
        else:
            params = new_avatar.param.values()
            del params['name']
            self._left_col[0].param.update(**params)

    def _render_reaction_icons(self):
        reaction_icons = self.reaction_icons
        reaction_icons.param.update(
            width=15,
            height=15,
            value=self.param.reactions,
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
            visible=self.param.show_reaction_icons
        )
        reaction_icons.link(self, value="reactions", bidirectional=True)
        return reaction_icons

    def _update_reaction_icons(self, _):
        self._icons_row[-1] = self._render_reaction_icons()

    def _update(self, ref, old_models):
        """
        Internals will be updated inplace.
        """

    def _update_object_pane(self, event=None):
        old = self._object_panel
        self._object_panel = new = self._create_panel(self.object, old=old)
        if old is not new:
            self._center_row[0] = new
        self._update_chat_copy_icon()

    @param.depends("avatar_lookup", "user", watch=True)
    def _update_avatar(self):
        """
        Update the avatar based on the user name.

        We do not use on_init here because if avatar is set,
        we don't want to override the provided avatar.

        However, if the user is updated, we want to update the avatar.
        """
        if self.avatar_lookup:
            self.avatar = self.avatar_lookup(self.user)
        else:
            self.avatar = avatar_lookup(
                self.user,
                self.avatar,
                self.default_avatars,
                DEFAULT_AVATARS,
            )

    def _update_chat_copy_icon(self):
        object_panel = self._object_panel
        if isinstance(object_panel, HTMLBasePane):
            object_panel = object_panel.object
        elif isinstance(object_panel, Widget):
            object_panel = object_panel.value
        if isinstance(object_panel, str) and self.show_copy_icon:
            self.chat_copy_icon.value = object_panel
            self.chat_copy_icon.visible = True
        else:
            self.chat_copy_icon.value = ""
            self.chat_copy_icon.visible = False

    def _cleanup(self, root=None) -> None:
        """
        Cleanup the exit stack.
        """
        self._composite._cleanup(root)
        return super()._cleanup(root)

    def stream(self, token: str, replace: bool = False):
        """
        Updates the message with the new token traversing the object to
        allow updating nested objects. When traversing a nested Panel
        the last object that supports rendering strings is updated, e.g.
        in a layout of `Column(Markdown(...), Image(...))` the Markdown
        pane is updated.

        Arguments
        ---------
        token: str
          The token to stream to the text pane.
        replace: bool (default=False)
            Whether to replace the existing text.
        """
        stream_to(obj=self.object, token=token, replace=replace, object_panel=self)

    def update(
        self,
        value: dict | ChatMessage | Any,
        user: str | None = None,
        avatar: str | bytes | BytesIO | None = None,
    ):
        """
        Updates the message with a new value, user and avatar.

        Arguments
        ---------
        value : ChatMessage | dict | Any
            The message contents to send.
        user : str | None
            The user to send as; overrides the message message's user if provided.
        avatar : str | bytes | BytesIO | None
            The avatar to use; overrides the message message's avatar if provided.
        """
        updates = {}
        if isinstance(value, dict):
            updates.update(value)
            if user:
                updates["user"] = user
            if avatar:
                updates["avatar"] = avatar
            if "value" in updates and "object" in updates:
                raise ValueError(
                    "Cannot set both 'value' and 'object' in the update dict."
                )
            updates["object"] = updates.pop("value", updates.get("object"))
        elif isinstance(value, ChatMessage):
            if user is not None or avatar is not None:
                raise ValueError(
                    "Cannot set user or avatar when explicitly sending "
                    "a ChatMessage. Set them directly on the ChatMessage."
                )
            updates = value.param.values()
        else:
            updates["object"] = value
        self.param.update(**updates)

    def select(
        self, selector: Optional[type | Callable[[Viewable], bool]] = None
    ) -> list[Viewable]:
        return super().select(selector) + self._composite.select(selector)

    def serialize(
        self,
        prefix_with_viewable_label: bool = True,
        prefix_with_container_label: bool = True
    ) -> str:
        """
        Format the object to a string.

        Arguments
        ---------
        prefix_with_viewable_label : bool
            Whether to include the name of the Viewable, or type
            of the viewable if no name is specified.
        prefix_with_container_label : bool
            Whether to include the name of the container, or type
            of the container if no name is specified.

        Returns
        -------
        str
            The serialized string.
        """
        return serialize_recursively(
            self.object,
            prefix_with_viewable_label=prefix_with_viewable_label,
            prefix_with_container_label=prefix_with_container_label,
        )



import pytest

from panel.chat.step import ChatStep


class TestChatStep:
    def test_initial_status(self):
        step = ChatStep()
        assert step.status == "pending", "Initial status should be 'pending'"

    def test_status_transitions_and_titles(self):
        step = ChatStep(running_title="Running", pending_title="Pending", success_title="Success", failed_title="Error Occurred")
        assert step.status == "pending", "Initial status should be 'pending'"

        # Test running status
        with step:
            assert step.status == "running", "Status should be 'running' during context execution"
            assert step.title == "Running", "Title should be 'Running' during context execution"

        # Test success status
        assert step.status == "success", "Status should be 'success' after context execution without errors"
        assert step.title == "Success", "Title should be 'Success' after context execution without errors"

        # Test failed status
        with pytest.raises(ValueError):
            with step:
                raise ValueError("Error")
        assert step.status == "failed", "Status should be 'failed' after an exception"
        assert step.title == "Error Occurred", "Title should update to 'Error Occurred' on failure"

    def test_avatar_and_collapsed_behavior(self):
        step = ChatStep(collapsed_on_success=True)
        initial_avatar = step._avatar_placeholder.object

        # Test avatar updates
        with step:
            assert step._avatar_placeholder.object != initial_avatar, "Avatar should change when status is 'running'"
        assert step._avatar_placeholder.object != initial_avatar, "Avatar should change when status is 'success'"

        # Test collapsed behavior
        assert step.collapsed, "Step should be collapsed on success when collapsed_on_success is True"

    def test_streaming_updates(self):
        step = ChatStep()

        # Test content streaming
        step.stream("First message")
        assert len(step.objects) == 1, "First message should be added to objects"
        assert step.objects[0].object == "First message", "First message should be 'First message'"
        step.stream("Second message")
        assert len(step.objects) == 1
        assert step.objects[0].object == "First messageSecond message", "Messages should be concatenated"

    def test_streaming_title_updates(self):
        step = ChatStep(running_title="Run")
        step.stream_title("Pend", status="pending")
        assert step.pending_title == "Pend", "Pending title should be 'Pend'"
        assert step.title == "Pend", "Title should be 'Pend' when status is 'pending'"

        step.stream_title("ing", status="pending")
        assert step.pending_title == "Pending", "Pending title should be 'Pending'"
        assert step.title == "Pending", "Title should be 'Pending' when status is 'pending'"

        step.stream_title("Starting now...", status="pending", replace=True)
        assert step.pending_title == "Starting now...", "Pending title should be 'Starting now...'"
        assert step.title == "Starting now...", "Title should be 'Starting now...' when status is 'pending'"

        step.stream_title("ning")
        assert step.running_title == "Running"
        assert step.title == "Starting now...", "Title should be 'Starting now...' when status is 'pending'"
        step.status = "running"
        assert step.title == "Running", "Title should be 'Running' when status is 'running'"

    def test_serialization(self):
        step = ChatStep()
        serialized = step.serialize()
        assert isinstance(serialized, str), "Serialization should return a string representation"

    def test_repeated_error(self):
        step = ChatStep()
        with pytest.raises(ValueError):
            with step:
                raise ValueError("Testing")

        assert step.status == "failed", "Status should be 'failed' after an exception"
        assert step.title == "Error: 'ValueError'", "Title should update to 'Error: 'ValueError'' on failure"
        assert step.objects[0].object == "Testing", "Error message should be streamed to the message pane"

        with pytest.raises(RuntimeError):
            with step:
                raise RuntimeError("Second Testing")

        assert step.status == "failed", "Status should be 'failed' after an exception"
        assert step.title == "Error: 'RuntimeError'", "Title should update to 'Error: 'RuntimeError'' on failure again"
        assert step.objects[0].object == "Testing\nSecond Testing", "Error message should be streamed to the message pane"

    def test_context_exception_ignore(self):
        step = ChatStep(context_exception="ignore")
        with step:
            raise ValueError("Testing")
        assert step.objects == []

    def test_context_exception_raise(self):
        step = ChatStep(context_exception="raise")
        with pytest.raises(ValueError, match="Testing"):
            with step:
                raise ValueError("Testing")
        assert step.objects[0].object == "Testing"

    def test_context_exception_summary(self):
        step = ChatStep(context_exception="summary")
        with step:
            raise ValueError("Testing")
        assert step.objects[0].object == "Testing"

    def test_context_exception_verbose(self):
        step = ChatStep(context_exception="verbose")
        with step:
            raise ValueError("Testing")
        assert "Traceback" in step.objects[0].object

    def test_stream_none(self):
        step = ChatStep()
        step.stream(None)
        assert len(step) == 1
        assert step[0].object == ""
        step.stream("abc")
        assert len(step) == 1
        assert step[0].object == "abc"

    def test_header_inherits_width(self):
        step = ChatStep(width=100)
        assert step.header.width == 100

    @pytest.mark.parametrize("width_key", ["max_width", "min_width"])
    def test_header_inherits_stretch_width(self, width_key):
        step = ChatStep(**{width_key: 100}, sizing_mode="stretch_width")
        assert getattr(step.header, width_key) == 100
        assert step.header.sizing_mode == "stretch_width"



import sys

import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel.tests.util import serve_component, wait_until
from panel.widgets import CodeEditor

pytestmark = pytest.mark.ui


def test_code_editor_on_keyup(page):

    editor = CodeEditor(value="print('Hello World!')", on_keyup=True)
    serve_component(page, editor)
    ace_input = page.locator(".ace_content")
    expect(ace_input).to_have_count(1)
    ace_input.click()

    page.keyboard.press("Enter")
    page.keyboard.type('print("Hello Panel!")')

    expect(page.locator(".ace_content")).to_have_text("print('Hello World!')\nprint(\"Hello Panel!\")", use_inner_text=True)
    wait_until(lambda: editor.value_input == "print('Hello World!')\nprint(\"Hello Panel!\")", page)
    assert editor.value == "print('Hello World!')\nprint(\"Hello Panel!\")"

    # clear the editor
    editor.value = ""
    expect(page.locator(".ace_content")).to_have_text("", use_inner_text=True)
    assert editor.value == ""
    assert editor.value_input == ""

    # enter Hello UI
    ace_input.click()
    page.keyboard.type('print("Hello UI!")')
    expect(page.locator(".ace_content")).to_have_text("print(\"Hello UI!\")", use_inner_text=True)

    wait_until(lambda: editor.value == "print(\"Hello UI!\")", page)


def test_code_editor_not_on_keyup(page):

    editor = CodeEditor(value="print('Hello World!')", on_keyup=False)
    serve_component(page, editor)
    ace_input = page.locator(".ace_content")
    expect(ace_input).to_have_count(1)
    ace_input.click()

    page.keyboard.press("Enter")
    page.keyboard.type('print("Hello Panel!")')

    expect(page.locator(".ace_content")).to_have_text("print('Hello World!')\nprint(\"Hello Panel!\")", use_inner_text=True)
    wait_until(lambda: editor.value_input == "print('Hello World!')\nprint(\"Hello Panel!\")")
    assert editor.value == "print('Hello World!')"

    # page click outside the editor; sync the value
    page.locator("body").click()
    assert editor.value_input == "print('Hello World!')\nprint(\"Hello Panel!\")"
    wait_until(lambda: editor.value == "print('Hello World!')\nprint(\"Hello Panel!\")")

    # clear the editor
    editor.value = ""
    expect(page.locator(".ace_content")).to_have_text("", use_inner_text=True)
    assert editor.value == ""
    assert editor.value_input == ""

    # enter Hello UI
    ace_input.click()
    page.keyboard.type('print("Hello UI!")')
    expect(page.locator(".ace_content")).to_have_text("print(\"Hello UI!\")", use_inner_text=True)
    assert editor.value == ""

    ctrl_key = 'Meta' if sys.platform == 'darwin' else 'Control'
    page.keyboard.down(ctrl_key)
    page.keyboard.press("Enter")
    page.keyboard.up(ctrl_key)

    wait_until(lambda: editor.value == "print(\"Hello UI!\")", page)



DEPENDENCY_NOT_INSTALLABLE = [
    "aiohttp", # https://github.com/aio-libs/aiohttp/issues/7253
    "datashader", # https://github.com/holoviz/datashader/issues/1200
    "pyarrow", # https://github.com/apache/arrow/issues/34996
    "pygraphviz", # https://github.com/pygraphviz/pygraphviz/issues/453
    "pyvista", # https://gitlab.kitware.com/vtk/vtk/-/issues/18806
    "streamz", # https://github.com/python-streamz/streamz/issues/467,
    "vtk", # https://gitlab.kitware.com/vtk/vtk/-/issues/18806
]

NOTEBOOK_ISSUES = {
    "gallery/vtk_interactive.ipynb": ["https://gitlab.kitware.com/vtk/vtk/-/issues/18806"],
    "gallery/vtk_slicer.ipynb": ["https://gitlab.kitware.com/vtk/vtk/-/issues/18806"],
    "gallery/vtk_warp.ipynb": ["https://gitlab.kitware.com/vtk/vtk/-/issues/18806"],
    "gallery/glaciers.ipynb": ["https://github.com/holoviz/panel/issues/4656"],
    "how_to/callbacks/async.md": ["https://github.com/aio-libs/aiohttp/issues/7253"],
    "reference/panes/DataFrame.ipynb": ["https://github.com/python-streamz/streamz/issues/467"],
    "reference/panes/IPyWidget.ipynb": ["https://github.com/holoviz/panel/issues/4394"],
    "reference/panes/Matplotlib.ipynb": ["https://github.com/holoviz/panel/issues/4394"],
    "reference/panes/Reacton.ipynb": ["https://github.com/holoviz/panel/issues/4394"],
    "reference/panes/Streamz.ipynb": ["https://github.com/python-streamz/streamz/issues/467"],
    "reference/panes/VTK.ipynb": ["https://gitlab.kitware.com/vtk/vtk/-/issues/18806"],
    "reference/panes/VTKVolume.ipynb": ["https://gitlab.kitware.com/vtk/vtk/-/issues/18806"],
    "reference/widgets/Debugger.ipynb": ["https://github.com/holoviz/panel/issues/4399"],
    "reference/widgets/EditableIntSlider.ipynb": ["https://github.com/holoviz/panel/issues/4400"],
    "reference/widgets/FileDownload.ipynb": ["https://github.com/holoviz/panel/issues/4401"],
    "reference/widgets/IntRangeSlider.ipynb": ["https://github.com/holoviz/panel/issues/4402"],
    "reference/widgets/MultiChoice.ipynb": ["https://github.com/holoviz/panel/issues/4403"],
    "reference/widgets/RangeSlider.ipynb": ["https://github.com/holoviz/panel/issues/4402"],
    "reference/widgets/SpeechToText.ipynb": ["https://github.com/holoviz/panel/issues/4404"],
    "reference/widgets/Terminal.ipynb": ["https://github.com/holoviz/panel/issues/4407"],
}



from datetime import date, datetime, time as dt_time
from pathlib import Path

import numpy as np
import pytest

from bokeh.models.widgets import FileInput as BkFileInput

from panel import config
from panel.widgets import (
    ArrayInput, Checkbox, DatePicker, DateRangePicker, DatetimeInput,
    DatetimePicker, DatetimeRangeInput, DatetimeRangePicker, FileInput,
    FloatInput, IntInput, LiteralInput, StaticText, TextInput, TimePicker,
)


def test_checkbox(document, comm):

    checkbox = Checkbox(value=True, name='Checkbox')

    widget = checkbox.get_root(document, comm=comm)

    assert isinstance(widget, checkbox._widget_type)
    assert widget.label == 'Checkbox'
    assert widget.active == True

    widget.active = False
    checkbox._process_events({'active': False})
    assert checkbox.value == False

    checkbox.value = True
    assert widget.active == True


def test_date_picker(document, comm):
    date_picker = DatePicker(name='DatePicker', value=date(2018, 9, 2),
                             start=date(2018, 9, 1), end=date(2018, 9, 10))

    widget = date_picker.get_root(document, comm=comm)

    assert isinstance(widget, date_picker._widget_type)
    assert widget.title == 'DatePicker'
    assert widget.value == '2018-09-02'
    assert widget.min_date == '2018-09-01'
    assert widget.max_date == '2018-09-10'

    widget.value = '2018-09-03'
    date_picker._process_events({'value': '2018-09-03'})
    assert date_picker.value == date(2018, 9, 3)

    date_picker._process_events({'value': date(2018, 9, 5)})
    assert date_picker.value == date(2018, 9, 5)

    date_picker._process_events({'value': date(2018, 9, 6)})
    assert date_picker.value == date(2018, 9, 6)

    date_picker.value = date(2018, 9, 4)
    assert widget.value == '2018-09-04'


def test_date_picker_options(document, comm):
    options = [date(2018, 9, 1), date(2018, 9, 2), date(2018, 9, 3)]
    datetime_picker = DatePicker(
        name='DatetimePicker', value=date(2018, 9, 2),
        options=options
    )
    assert datetime_picker.value == date(2018, 9, 2)
    assert datetime_picker.enabled_dates == options


def test_daterange_picker(document, comm):
    date_range_picker = DateRangePicker(name='DateRangePicker',
                                        value=(date(2018, 9, 2), date(2018, 9, 3)),
                                        start=date(2018, 9, 1),
                                        end=date(2018, 9, 10))

    widget = date_range_picker.get_root(document, comm=comm)

    assert isinstance(widget, date_range_picker._widget_type)
    assert widget.title == 'DateRangePicker'
    assert widget.value == ('2018-09-02', '2018-09-03')
    assert widget.min_date == '2018-09-01'
    assert widget.max_date == '2018-09-10'

    date_range_picker._process_events({'value': ('2018-09-03', '2018-09-04')})
    assert date_range_picker.value == (date(2018, 9, 3), date(2018, 9, 4))

    date_range_picker._process_events({'value': ('2018-09-05', '2018-09-08')})
    assert date_range_picker.value == (date(2018, 9, 5), date(2018, 9, 8))

    value = date_range_picker._process_param_change({'value': (date(2018, 9, 4), date(2018, 9, 5))})
    assert value['value'] == ('2018-09-04', '2018-09-05')

    value = date(2018, 9, 4)
    assert date_range_picker._convert_date_to_string(value) == '2018-09-04'
    assert date_range_picker._convert_string_to_date(date_range_picker._convert_date_to_string(value)) == value

    # Check start value
    with pytest.raises(ValueError):
        date_range_picker._process_events({'value': ('2018-08-31', '2018-09-01')})

    # Check end value
    with pytest.raises(ValueError):
        date_range_picker._process_events({'value': ('2018-09-10', '2018-09-11')})


def test_datetime_picker(document, comm):
    datetime_picker = DatetimePicker(
        name='DatetimePicker', value=datetime(2018, 9, 2, 1, 5),
        start=date(2018, 9, 1), end=datetime(2018, 9, 10)
    )

    widget = datetime_picker.get_root(document, comm=comm)

    assert isinstance(widget, datetime_picker._widget_type)
    assert widget.title == 'DatetimePicker'
    assert widget.value == '2018-09-02 01:05:00'
    assert widget.min_date == '2018-09-01T00:00:00'
    assert widget.max_date == '2018-09-10T00:00:00'

    datetime_picker._process_events({'value': '2018-09-03 03:00:01'})
    assert datetime_picker.value == datetime(2018, 9, 3, 3, 0, 1)

    datetime_picker._process_events({'value': datetime(2018, 9, 5)})
    assert datetime_picker.value == datetime(2018, 9, 5)

    value = datetime_picker._process_param_change({'value': datetime(2018, 9, 4, 1, 0, 1)})
    assert value['value'] == '2018-09-04 01:00:01'

    value = datetime(2018, 9, 4, 12, 1)
    assert datetime_picker._deserialize_value(value) == '2018-09-04 12:01:00'
    assert datetime_picker._serialize_value(datetime_picker._deserialize_value(value)) == value

    # Check start value
    with pytest.raises(ValueError):
        datetime_picker._process_events({'value': '2018-08-31 23:59:59'})

    # Check end value
    with pytest.raises(ValueError):
        datetime_picker._process_events({'value': '2018-09-10 00:00:01'})


def test_datetime_picker_options(document, comm):
    options = [datetime(2018, 9, 1), datetime(2018, 9, 2), datetime(2018, 9, 3)]
    datetime_picker = DatetimePicker(
        name='DatetimePicker', value=datetime(2018, 9, 2, 1, 5),
        options=options
    )
    assert datetime_picker.value == datetime(2018, 9, 2, 1, 5)
    assert datetime_picker.enabled_dates == options


def test_datetime_range_picker(document, comm):
    datetime_range_picker = DatetimeRangePicker(
        name='DatetimeRangePicker', value=(datetime(2018, 9, 2, 1, 5), datetime(2018, 9, 2, 1, 6)),
        start=date(2018, 9, 1), end=datetime(2018, 9, 10)
    )

    widget = datetime_range_picker.get_root(document, comm=comm)

    assert isinstance(widget, datetime_range_picker._widget_type)
    assert widget.title == 'DatetimeRangePicker'
    assert widget.value == '2018-09-02 01:05:00 to 2018-09-02 01:06:00'
    assert widget.min_date == '2018-09-01T00:00:00'
    assert widget.max_date == '2018-09-10T00:00:00'

    datetime_range_picker._process_events({'value': '2018-09-03 03:00:01 to 2018-09-04 03:00:01'})
    assert datetime_range_picker.value == (datetime(2018, 9, 3, 3, 0, 1), datetime(2018, 9, 4, 3, 0, 1))

    value = datetime_range_picker._process_param_change(
        {'value': (datetime(2018, 9, 4, 1, 0, 1), datetime(2018, 9, 4, 4, 0, 1))}
    )
    assert value['value'] == '2018-09-04 01:00:01 to 2018-09-04 04:00:01'

    value = (datetime(2018, 9, 4, 12, 1), datetime(2018, 9, 4, 12, 1, 10))
    assert datetime_range_picker._deserialize_value(value) == '2018-09-04 12:01:00 to 2018-09-04 12:01:10'
    assert datetime_range_picker._serialize_value(datetime_range_picker._deserialize_value(value)) == value

    # Check start value
    with pytest.raises(ValueError):
        datetime_range_picker._process_events({'value': '2018-08-31 23:59:59'})

    # Check end value
    with pytest.raises(ValueError):
        datetime_range_picker._process_events({'value': '2018-09-10 00:00:01'})


def test_time_picker(document, comm):
    time_picker = TimePicker(name='Time Picker', value=dt_time(hour=18), format='H:i K')
    assert time_picker.value == dt_time(hour=18)
    assert time_picker.format == 'H:i K'
    assert time_picker.start is None
    assert time_picker.end is None


def test_time_picker_str(document, comm):
    time_picker = TimePicker(name='Time Picker', value="08:28", start='00:00', end='12:00')
    assert time_picker.value == "08:28"
    assert time_picker.format == 'H:i'
    assert time_picker.start == "00:00"
    assert time_picker.end == "12:00"


def test_file_input(document, comm):
    file_input = FileInput(accept='.txt')

    widget = file_input.get_root(document, comm=comm)

    assert isinstance(widget, BkFileInput)

    file_input._process_events({'mime_type': 'text/plain', 'value': 'U29tZSB0ZXh0Cg==', 'filename': 'testfile'})
    assert file_input.value == b'Some text\n'
    assert file_input.mime_type == 'text/plain'
    assert file_input.accept == '.txt'
    assert file_input.filename == 'testfile'


def test_literal_input(document, comm):

    literal = LiteralInput(value={}, type=dict, name='Literal')

    widget = literal.get_root(document, comm=comm)

    assert isinstance(widget, literal._widget_type)
    assert widget.title == 'Literal'
    assert widget.value == '{}'

    literal._process_events({'value': "{'key': (0, 2)}"})
    assert literal.value == {'key': (0, 2)}
    assert widget.title == 'Literal'

    literal._process_events({'value': "(0, 2)"})
    assert literal.value == {'key': (0, 2)}
    assert widget.title == 'Literal (wrong type)'

    literal._process_events({'value': "invalid"})
    assert literal.value == {'key': (0, 2)}
    assert widget.title == 'Literal (invalid)'

    literal._process_events({'value': "{'key': (0, 3)}"})
    assert literal.value == {'key': (0, 3)}
    assert widget.title == 'Literal'

    with pytest.raises(ValueError):
        literal.value = []

def test_static_text(document, comm):

    text = StaticText(value='ABC', name='Text:')

    widget = text.get_root(document, comm=comm)

    assert isinstance(widget, text._widget_type)
    assert widget.text == '<b>Text:</b>: ABC'

    text.value = 'CBA'
    assert widget.text == '<b>Text:</b>: CBA'

    text.value = '<b>Text:</b>: ABC'
    assert widget.text == '<b>Text:</b>: ABC'

def test_static_text_no_sync(document, comm):
    text = StaticText(value='ABC', name='Text:')

    widget = text.get_root(document, comm=comm)

    widget.text = 'CBA'
    assert text.value == 'ABC'

def test_static_text_empty(document, comm):

    text = StaticText(name='Text:')

    widget = text.get_root(document, comm=comm)

    assert widget.text == '<b>Text:</b>: '

def test_static_text_repr(document, comm):

    text = StaticText(value=StaticText, name='Text:')

    widget = text.get_root(document, comm=comm)

    assert widget.text == '<b>Text:</b>: &lt;class &#x27;panel.widgets.input.StaticText&#x27;&gt;'



def test_text_input(document, comm):

    text = TextInput(value='ABC', name='Text:')

    widget = text.get_root(document, comm=comm)

    assert isinstance(widget, text._widget_type)
    assert widget.value == 'ABC'
    assert widget.title == 'Text:'

    text._process_events({'value': 'CBA'})
    assert text.value == 'CBA'

    text.value = 'A'
    assert widget.value == 'A'

def test_datetime_input(document, comm):
    dt_input = DatetimeInput(value=datetime(2018, 1, 1),
                             start=datetime(2017, 12, 31),
                             end=datetime(2018, 1, 10),
                             name='Datetime')

    widget = dt_input.get_root(document, comm=comm)

    assert isinstance(widget, dt_input._widget_type)
    assert widget.title == 'Datetime'
    assert widget.value == '2018-01-01 00:00:00'

    dt_input._process_events({'value': '2018-01-01 00:00:01'})
    assert dt_input.value == datetime(2018, 1, 1, 0, 0, 1)
    assert widget.title == 'Datetime'

    dt_input._process_events({'value': '2018-01-01 00:00:01a'})
    assert dt_input.value == datetime(2018, 1, 1, 0, 0, 1)
    assert widget.title == 'Datetime (invalid)'

    dt_input._process_events({'value': '2018-01-11 00:00:00'})
    assert dt_input.value == datetime(2018, 1, 1, 0, 0, 1)
    assert widget.title == 'Datetime (out of bounds)'

    dt_input._process_events({'value': '2018-01-02 00:00:01'})
    assert dt_input.value == datetime(2018, 1, 2, 0, 0, 1)
    assert widget.title == 'Datetime'

    with pytest.raises(ValueError):
        dt_input.value = datetime(2017, 12, 30)


def test_datetime_range_input(document, comm):
    dt_input = DatetimeRangeInput(
        value=(datetime(2018, 1, 1), datetime(2018, 1, 3)),
        start=datetime(2017, 12, 31), end=datetime(2018, 1, 10),
        name='Datetime')

    composite = dt_input.get_root(document, comm=comm)
    label, start, end = composite.children

    assert isinstance(composite, dt_input._composite_type._bokeh_model)
    assert label.text == 'Datetime'
    assert start.value == '2018-01-01 00:00:00'
    assert end.value == '2018-01-03 00:00:00'

    dt_input._start._process_events({'value': '2018-01-01 00:00:01'})
    assert dt_input.value == (datetime(2018, 1, 1, 0, 0, 1), datetime(2018, 1, 3))
    assert label.text == 'Datetime'

    dt_input._start._process_events({'value': '2018-01-01 00:00:01a'})
    assert dt_input.value == (datetime(2018, 1, 1, 0, 0, 1), datetime(2018, 1, 3))
    assert label.text == 'Datetime (invalid)'

    dt_input._start._process_events({'value': '2018-01-11 00:00:00'})
    assert dt_input.value == (datetime(2018, 1, 1, 0, 0, 1), datetime(2018, 1, 3))
    assert label.text == 'Datetime (out of bounds)'

    dt_input._end._process_events({'value': '2018-01-11 00:00:00a'})
    assert dt_input.value == (datetime(2018, 1, 1, 0, 0, 1), datetime(2018, 1, 3))
    assert label.text == 'Datetime (out of bounds) (invalid)'

    dt_input._start._process_events({'value': '2018-01-02 00:00:00'})
    dt_input._end._process_events({'value': '2018-01-03 00:00:00'})
    assert dt_input.value == (datetime(2018, 1, 2), datetime(2018, 1, 3))
    assert label.text == 'Datetime'

    dt_input._start._process_events({'value': '2018-01-05 00:00:00'})
    assert dt_input.value == (datetime(2018, 1, 2), datetime(2018, 1, 3))
    assert label.text == 'Datetime (start of range must be <= end)'

    dt_input._start._process_events({'value': '2018-01-01 00:00:00'})
    assert dt_input.value == (datetime(2018, 1, 1), datetime(2018, 1, 3))
    assert label.text == 'Datetime'


def test_int_input(document, comm):
    int_input = IntInput(name='Int input')
    widget = int_input.get_root(document, comm=comm)

    assert isinstance(widget, int_input._widget_type)
    assert widget.title == 'Int input'
    assert widget.step == 1
    assert widget.value == 0

    int_input._process_events({'value': 2})
    assert int_input.value == 2
    int_input._process_events({'value_throttled': 2})
    assert int_input.value_throttled == 2

    int_input.value = 0
    assert widget.value == 0

    # Testing throttled mode
    with config.set(throttled=True):
        int_input._process_events({'value': 1})
        assert int_input.value == 0  # no change
        int_input._process_events({'value_throttled': 1})
        assert int_input.value == 1

        int_input.value = 2
        assert widget.value == 2


def test_float_input(document, comm):
    float_input = FloatInput(value=0.4, name="Float input")
    widget = float_input.get_root(document, comm=comm)

    assert isinstance(widget, float_input._widget_type)
    assert widget.title == 'Float input'
    assert widget.step == 0.1
    assert widget.value == 0.4

    float_input._process_events({'value': 0.2})
    assert float_input.value == 0.2
    float_input._process_events({'value_throttled': 0.2})
    assert float_input.value_throttled == 0.2

    float_input.value = 0.3
    assert widget.value == 0.3

    # Testing throttled mode
    with config.set(throttled=True):
        float_input._process_events({'value': 0.4})
        assert float_input.value == 0.3  # no change
        float_input._process_events({'value_throttled': 0.4})
        assert float_input.value == 0.4

        float_input.value = 0.5
        assert widget.value == 0.5


def test_array_input(document, comm):
    array_input = ArrayInput(value=np.array([1, 2, 3]),  name="Array input", max_array_size=3)
    widget = array_input.get_root(document, comm=comm)

    assert isinstance(widget, array_input._widget_type)
    assert widget.title == 'Array input'
    assert widget.value == '[1, 2, 3]'
    assert not widget.disabled

    array_input.value = np.array([1, 2, 3, 4])

    assert widget.value == "'[1,2,3,4]'"
    assert widget.disabled
    assert array_input._auto_disabled

    array_input.value = np.array([1, 2, 3])
    assert not widget.disabled
    assert not array_input._auto_disabled


def test_file_input_save_one_file(document, comm, tmpdir):
    file_input = FileInput(accept='.txt')

    widget = file_input.get_root(document, comm=comm)

    assert isinstance(widget, BkFileInput)

    file_input._process_events({'mime_type': 'text/plain', 'value': 'U29tZSB0ZXh0Cg==', 'filename': 'testfile'})

    fpath = Path(tmpdir) / 'out.txt'
    file_input.save(str(fpath))

    assert fpath.exists()
    content = fpath.read_text()
    assert content == 'Some text\n'



from typing import Any

from bokeh.core.properties import List, String
from bokeh.events import ModelEvent

from .layout import Column


class ScrollLatestEvent(ModelEvent):

    event_name = 'scroll_latest_event'

    def __init__(self, model, rerender=False):
        super().__init__(model=model)
        self.rerender = rerender

    def event_values(self) -> dict[str, Any]:
        return dict(super().event_values(), rerender=self.rerender)


class ScrollButtonClick(ModelEvent):

    event_name = 'scroll_button_click'

    def __init__(self, model, data=None):
        self.data = data
        super().__init__(model=model)


class Feed(Column):

    visible_children = List(String())



"""
Defines various Select widgets which allow choosing one or more items
from a list of options.
"""
from __future__ import annotations

import itertools
import re

from functools import partial
from types import FunctionType
from typing import (
    TYPE_CHECKING, Any, Awaitable, Callable, ClassVar, Mapping, Optional,
)

import numpy as np
import param

from bokeh.models import PaletteSelect
from bokeh.models.widgets import (
    AutocompleteInput as _BkAutocompleteInput,
    CheckboxGroup as _BkCheckboxGroup, MultiChoice as _BkMultiChoice,
    RadioGroup as _BkRadioBoxGroup,
)

from ..io.resources import CDN_DIST
from ..io.state import state
from ..layout.base import Column, ListPanel, NamedListPanel
from ..models import (
    CheckboxButtonGroup as _BkCheckboxButtonGroup,
    CustomMultiSelect as _BkMultiSelect, CustomSelect,
    RadioButtonGroup as _BkRadioButtonGroup, SingleSelect as _BkSingleSelect,
)
from ..util import PARAM_NAME_PATTERN, indexOf, isIn
from ._mixin import TooltipMixin
from .base import CompositeWidget, Widget
from .button import Button, _ButtonBase
from .input import TextAreaInput, TextInput

if TYPE_CHECKING:
    from bokeh.document import Document
    from bokeh.model import Model
    from pyviz_comms import Comm

    from ..models.widgets import DoubleClickEvent


class SelectBase(Widget):

    options = param.ClassSelector(default=[], class_=(dict, list))

    __abstract = True

    @property
    def labels(self):
        labels = []
        for o in self.options:
            if isinstance(o, param.Parameterized) and not PARAM_NAME_PATTERN.match(o.name):
                labels.append(o.name)
            else:
                labels.append(str(o))
        return labels

    @property
    def values(self):
        if isinstance(self.options, dict):
            return list(self.options.values())
        else:
            return self.options

    @property
    def _items(self):
        return dict(zip(self.labels, self.values))


class SingleSelectBase(SelectBase):

    value = param.Parameter(default=None)

    _allows_values: ClassVar[bool] = True

    _allows_none: ClassVar[bool] = False

    _supports_embed: ClassVar[bool] = True

    __abstract = True

    def __init__(self, **params):
        super().__init__(**params)
        values = self.values
        if self.value is None and None not in values and values and not self._allows_none:
            self.value = values[0]

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        labels, values = self.labels, self.values
        unique = len(set(self.unicode_values)) == len(labels) and self._allows_values
        if 'value' in msg:
            val = msg['value']
            if isIn(val, values):
                unicode_values = self.unicode_values if unique else labels
                msg['value'] = unicode_values[indexOf(val, values)]
            elif values:
                self.value = self.param['value'].default if self._allows_none else self.values[0]
                if not self._allows_none:
                    del msg['value']
            else:
                self.value = self.param['value'].default
                if self._allows_none:
                    msg['value'] = self.value

        option_prop = self._property_mapping.get('options', 'options')
        is_list = isinstance(self.param['value'], param.List)
        if option_prop in msg and not is_list:
            if isinstance(self.options, dict):
                if unique and self._allows_values:
                    options = [(v, l) for l,v in zip(labels, self.unicode_values)]
                else:
                    options = labels
                msg[option_prop] = options
            else:
                msg[option_prop] = self.unicode_values
            val = self.value
            if values:
                if not isIn(val, values):
                    self.value = self.param['value'].default if self._allows_none else values[0]
            else:
                self.value = self.param['value'].default
        return msg

    @property
    def unicode_values(self):
        return [str(v) for v in self.values]

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        if 'value' in msg:
            if not self.values:
                pass
            elif msg['value'] == '':
                msg['value'] = self.values[0] if self.values else None
            else:
                if isIn(msg['value'], self.unicode_values):
                    idx = indexOf(msg['value'], self.unicode_values)
                else:
                    idx = indexOf(msg['value'], self.labels)
                msg['value'] = self._items[self.labels[idx]]
        msg.pop('options', None)
        return msg

    def _get_embed_state(self, root, values=None, max_opts=3):
        if values is None:
            values = self.values
        elif any(v not in self.values for v in values):
            raise ValueError("Supplied embed states were not found "
                             f"in the {type(self).__name__} widgets values list.")
        return (self, self._models[root.ref['id']][0], values,
                lambda x: x.value, 'value', 'cb_obj.value')


class Select(SingleSelectBase):
    """
    The `Select` widget allows selecting a value from a list or dictionary of
    `options` by selecting it from a dropdown menu or selection area.

    It falls into the broad category of single-value, option-selection widgets
    that provide a compatible API and include the `RadioBoxGroup`,
    `AutocompleteInput` and `DiscreteSlider` widgets.

    Reference: https://panel.holoviz.org/reference/widgets/Select.html

    :Example:

    >>> Select(name='Study', options=['Biology', 'Chemistry', 'Physics'])
    """

    description = param.String(default=None, doc="""
        An HTML string describing the function of this component.""")

    disabled_options = param.List(default=[], nested_refs=True, doc="""
        Optional list of ``options`` that are disabled, i.e. unusable and
        un-clickable. If ``options`` is a dictionary the list items must be
        dictionary values.""")

    groups = param.Dict(default=None, nested_refs=True, doc="""
        Dictionary whose keys are used to visually group the options
        and whose values are either a list or a dictionary of options
        to select from. Mutually exclusive with ``options``  and valid only
        if ``size`` is 1.""")

    size = param.Integer(default=1, bounds=(1, None), doc="""
        Declares how many options are displayed at the same time.
        If set to 1 displays options as dropdown otherwise displays
        scrollable area.""")

    width = param.Integer(default=300, allow_None=True, doc="""
      Width of this component. If sizing_mode is set to stretch
      or scale mode this will merely be used as a suggestion.""")

    _rename: ClassVar[Mapping[str, str | None]] = {
        'groups': None,
    }

    _source_transforms: ClassVar[Mapping[str, str | None]] = {
        'size': None, 'groups': None
    }

    _stylesheets: ClassVar[list[str]] = [f'{CDN_DIST}css/select.css']

    @property
    def _widget_type(self):
        return CustomSelect if self.size == 1 else _BkSingleSelect

    def __init__(self, **params):
        super().__init__(**params)
        if self.size == 1:
            self.param.size.constant = True
        self._internal_callbacks.extend([
            self.param.watch(
                self._validate_options_groups,
                ['options', 'groups']
            ),
            self.param.watch(
                self._validate_disabled_options,
                ['options', 'disabled_options', 'value']
            ),
        ])
        self._validate_options_groups()
        self._validate_disabled_options()

    def _validate_disabled_options(self, *events):
        if self.disabled_options and self.disabled_options == self.values:
            raise ValueError(
                f'All the options of a {type(self).__name__} '
                'widget cannot be disabled.'
            )
        not_in_opts = [
            dopts
            for dopts in self.disabled_options
            if dopts not in (self.values or [])
        ]
        if not_in_opts:
            raise ValueError(
                f'Cannot disable non existing options of {type(self).__name__}: {not_in_opts}'
            )
        if len(events) == 1:
            if events[0].name == 'value' and self.value in self.disabled_options:
                raise ValueError(
                    f'Cannot set the value of {type(self).__name__} to '
                    f'{self.value!r} as it is a disabled option.'
                )
            elif events[0].name == 'disabled_options' and self.value in self.disabled_options:
                raise ValueError(
                    f'Cannot set disabled_options of {type(self).__name__} to a list that '
                    f'includes the current value {self.value!r}.'
                )
        if self.value in self.disabled_options:
            raise ValueError(
                f'Cannot initialize {type(self).__name__} with value {self.value!r} '
                'as it is one of the disabled options.'
            )

    def _validate_options_groups(self, *events):
        if self.options and self.groups:
            raise ValueError(
                f'{type(self).__name__} options and groups parameters '
                'are mutually exclusive.'
            )
        if self.size > 1 and self.groups:
            raise ValueError(
                f'{type(self).__name__} with size > 1 doe not support the'
                ' `groups` parameter, use `options` instead.'
            )

    def _process_param_change(self, msg: dict[str, Any]) -> dict[str, Any]:
        groups_provided = 'groups' in msg
        msg = super()._process_param_change(msg)
        if groups_provided or 'options' in msg and self.groups:
            groups = self.groups
            if (all(isinstance(values, dict) for values in groups.values()) is False
               and  all(isinstance(values, list) for values in groups.values()) is False):
                raise ValueError(
                    'The values of the groups dictionary must be all of '
                    'the dictionary or the list type.'
                )
            labels, values = self.labels, self.values
            unique = len(set(self.unicode_values)) == len(labels)
            if groups:
                if isinstance(next(iter(self.groups.values())), dict):
                    if unique:
                        options = {
                            group: [(str(value), label) for label, value in subd.items()]
                            for group, subd in groups.items()
                        }
                    else:
                        options = {
                            group: [str(v) for v in self.groups[group]]
                            for group in groups.keys()
                        }
                    msg['options'] = options
                else:
                    msg['options'] = {
                        group: [(str(value), str(value)) for value in values]
                        for group, values in groups.items()
                    }
            val = self.value
            if values:
                if not isIn(val, values):
                    self.value = values[0]
            else:
                self.value = None
        return msg

    @property
    def labels(self):
        if self.options:
            return super().labels
        else:
            if not self.groups:
                return {}
            else:
                return list(map(str, itertools.chain(*self.groups.values())))

    @property
    def values(self):
        if self.options:
            return super().values
        else:
            if not self.groups:
                return []
            if isinstance(next(iter(self.groups.values())), dict):
                return [v for subd in self.groups.values() for v in subd.values()]
            else:
                return list(itertools.chain(*self.groups.values()))


class NestedSelect(CompositeWidget):
    """
    The `NestedSelect` widget is composed of multiple widgets, where subsequent select options
    depend on the parent's value.

    Reference: https://panel.holoviz.org/reference/widgets/NestedSelect.html

    :Example:

    >>> NestedSelect(
    ...     options={
    ...         "gfs": {"tmp": [1000, 500], "pcp": [1000]},
    ...         "name": {"tmp": [1000, 925, 850, 700, 500], "pcp": [1000]},
    ...     },
    ...     levels=["model", "var", "level"],
    ... )
    """

    value = param.Dict(doc="""
        The value from all the Select widgets; the keys are the levels names.
        If no levels names are specified, the keys are the levels indices.""")

    options = param.ClassSelector(class_=(dict, FunctionType), doc="""
        The options to select from. The options may be nested dictionaries, lists,
        or callables that return those types. If callables are used, the callables
        must accept `level` and `value` keyword arguments, where `level` is the
        level that updated and `value` is a dictionary of the current values, containing keys
        up to the level that was updated.""")

    layout = param.Parameter(default=Column, doc="""
        The layout type of the widgets. If a dictionary, a "type" key can be provided,
        to specify the layout type of the widgets, and any additional keyword arguments
        will be used to instantiate the layout.""")

    levels = param.List(doc="""
        Either a list of strings or a list of dictionaries. If a list of strings, the strings
        are used as the names of the levels. If a list of dictionaries, each dictionary may
        have a "name" key, which is used as the name of the level, a "type" key, which
        is used as the type of widget, and any corresponding widget keyword arguments.
        Must be specified if options is callable.""")

    disabled = param.Boolean(default=False, doc="""
        Whether the widget is disabled.""")

    _widgets = param.List(doc="The nested select widgets.")

    _max_depth = param.Integer(doc="The number of levels of the nested select widgets.")

    _levels = param.List(doc="""
        The internal rep of levels to prevent overwriting user provided levels.""")

    def __init__(self, **params):
        super().__init__(**params)
        self._update_widgets()

    def _gather_values_from_widgets(self, up_to_i=None):
        """
        Gather values from all the select widgets to update the class' value.
        """
        values = {}
        for i, select in enumerate(self._widgets):
            if up_to_i is not None and i >= up_to_i:
                break
            level = self._levels[i]
            if isinstance(level, dict):
                name = level.get("name", i)
            else:
                name = level
            values[name] = select.value if select.options else None

        return values

    def _uses_callable(self, d):
        """
        Check if the nested options has a callable.
        """
        if callable(d):
            return True

        if isinstance(d, dict):
            for value in d.values():
                if callable(value):
                    return True
                elif isinstance(value, dict):
                    return self._uses_callable(value)
        return False

    def _find_max_depth(self, d, depth=1):
        if isinstance(d, list) or d is None:
            return depth-1
        max_depth = depth
        for value in d.values():
            max_depth = max(max_depth, self._find_max_depth(value, depth + 1))
        return max_depth

    def _resolve_callable_options(self, i, options) -> dict | list:
        level = self.levels[i]
        value = self._gather_values_from_widgets(up_to_i=i)
        options = options(level=level, value=value)
        return options

    @param.depends("options", "layout", "levels", watch=True)
    def _update_widgets(self):
        """
        When options is changed, reflect changes on the select widgets.
        """
        if self._uses_callable(self.options):
            if not self.levels:
                raise ValueError("levels must be specified if options is callable")
            self._max_depth = len(self.levels)
        else:
            self._max_depth = self._find_max_depth(self.options) + 1

        if not self.levels:
            self._levels = [i for i in range(self._max_depth)]
        elif len(self.levels) != self._max_depth:
            raise ValueError(f"levels must be of length {self._max_depth}")
        else:
            self._levels = self.levels

        self._widgets = []

        # use [] as default because it's the last level if options is None
        options = (self.options or [])
        if isinstance(self.options, dict):
            options = self.options.copy()

        for i in range(self._max_depth):
            if callable(options):
                options = self._resolve_callable_options(i, options)

            value = self._init_widget(i, options)
            if isinstance(options, dict) and len(options) > 0 and value is not None:
                options = options[value]
            elif i < self._max_depth - 1 and not isinstance(options, dict):
                raise ValueError(
                    f"The level, {self.levels[i]!r} is not the last nested level, "
                    f"so it must be a dict, but got {options!r}, which is a "
                    f"{type(options).__name__}"
                )

        if isinstance(self.layout, dict):
            layout_type = self.layout.pop("type", Column)
            layout_kwargs = self.layout.copy()
        elif issubclass(self.layout, (ListPanel, NamedListPanel)):
            layout_type = self.layout
            layout_kwargs = {}
        else:
            raise ValueError(
                f"The layout must be a subclass of ListLike or dict, got {self.layout!r}."
            )

        self._composite = layout_type(*self._widgets, **layout_kwargs)
        if self.options is not None:
            self.value = self._gather_values_from_widgets()

    def _extract_level_metadata(self, i):
        """
        Extract the widget type and keyword arguments from the level metadata.
        """
        level = self._levels[i]
        if isinstance(level, int):
            return Select, {}
        elif isinstance(level, str):
            return Select, {"name": level}
        widget_type = level.get("type", Select)
        widget_kwargs = {k: v for k, v in level.items() if k != "type"}
        return widget_type, widget_kwargs

    def _lookup_value(self, i, options, values, name=None, error=False):
        """
        Look up the value of the select widget at index i or by name.
        """
        options_iterable = isinstance(options, (list, dict))
        if values is None or (options_iterable and len(options) == 0):
            value = None
        elif name is None:
            # get by index
            value = list(values.values())[i] if i < len(values) else None
        elif isinstance(self._levels[0], int):
            # get by levels keys, which are enumerations
            value = values.get(i)
        else:
            # get by levels keys, which are strings
            value = values.get(name)

        if options_iterable and options and value not in options:
            if value is not None and error:
                raise ValueError(
                    f"Failed to set value {value!r} for level {name!r}, "
                    f"must be one of {options!r}."
                )
            else:
                value = options[0]
        return value

    def _init_widget(self, i, options):
        """
        Helper method to initialize a select widget.
        """
        if isinstance(options, dict):
            options = list(options.keys())
        elif not isinstance(options, (list, dict)) and not callable(options):
            raise ValueError(
                f"options must be a dict, list, or callable that returns those types, "
                f"got {options!r}, which is a {type(options).__name__}"
            )

        widget_type, widget_kwargs = self._extract_level_metadata(i)
        value = self._lookup_value(i, options, self.value, error=False)
        widget_kwargs["options"] = options
        widget_kwargs["value"] = value
        if "visible" not in widget_kwargs:
            # first select widget always visible
            widget_kwargs["visible"] = i == 0 or callable(options) or len(options) > 0
        widget = widget_type(**widget_kwargs)
        self.link(widget, disabled="disabled")
        widget.param.watch(self._update_widget_options_interactively, "value")
        self._widgets.append(widget)
        return value

    def _update_widget_options_interactively(self, event):
        """
        When a select widget's value is changed, update to the latest options.
        """
        if self.options is None:
            return

        # little optimization to avoid looping through all the
        # widgets and updating their value
        for start_i, select in enumerate(self._widgets):  # noqa: B007
            if select is event.obj:
                break

        options = self.options if callable(self.options) else self.options.copy()

        # batch watch to prevent continuously triggering
        # this function when updating the select widgets
        with param.parameterized.batch_call_watchers(self):
            for i, select in enumerate(self._widgets[:-1]):
                if select.value is None:
                    options = {}
                    visible = False
                elif options:
                    if isinstance(options, dict):
                        if select.value in options:
                            options = options[select.value]
                        else:
                            options = options[list(options.keys())[0]]
                    visible = bool(options)

                if i < start_i:
                    # If the select widget is before the one
                    # that triggered the event,
                    # then we don't need to update it;
                    # we just need to subset options.
                    continue

                next_select = self._widgets[i + 1]
                if callable(options):
                    options = self._resolve_callable_options(i + 1, options)
                    next_options = list(options)
                elif isinstance(options, dict):
                    next_options = list(options.keys())
                elif isinstance(options, list):
                    next_options = options
                else:
                    raise NotImplementedError(
                        "options must be a dict, list, or callable that returns those types."
                    )

                next_select.param.update(
                    options=next_options,
                    visible=visible
                )
            self.value = self._gather_values_from_widgets()

    @param.depends("value", watch=True)
    def _update_options_programmatically(self):
        """
        When value is passed, update to the latest options.
        """
        if self.options is None:
            return

        # must define these or else it gets mutated in the loop
        options = self.options if callable(self.options) else self.options.copy()
        set_values = self.value.copy()
        original_values = self._gather_values_from_widgets()

        if set_values == original_values:
            return

        with param.parameterized.batch_call_watchers(self):
            try:
                for i in range(self._max_depth):
                    curr_select = self._widgets[i]
                    if callable(options):
                        options = self._resolve_callable_options(i, options)
                        curr_options = list(options)
                    elif isinstance(options, dict):
                        curr_options = list(options.keys())
                    else:
                        curr_options = options
                    curr_value = self._lookup_value(
                        i, curr_options, set_values,
                        name=curr_select.name, error=True
                    )

                    with param.discard_events(self):
                        curr_select.param.update(
                            options=curr_options,
                            value=curr_value,
                            visible=callable(curr_options) or len(curr_options) > 0
                        )
                    if curr_value is None:
                        break
                    if i < self._max_depth - 1:
                        options = options[curr_value]
            except Exception:
                # revert to original values if there is an error
                # so it's not in a limbo state
                self.value = original_values
                raise


class ColorMap(SingleSelectBase):
    """
    The `ColorMap` widget allows selecting a value from a dictionary of
    `options` each containing a colormap specified as a list of colors
    or a matplotlib colormap.

    Reference: https://panel.holoviz.org/reference/widgets/ColorMap.html

    :Example:

    >>> ColorMap(name='Reds', options={'Reds': ['white', 'red'], 'Blues': ['#ffffff', '#0000ff']})
    """

    options = param.Dict(default={}, doc="""
        Dictionary of colormaps""")

    ncols = param.Integer(default=1, doc="""
        Number of columns of swatches to display.""")

    swatch_height = param.Integer(default=20, doc="""
        Height of the color swatches.""")

    swatch_width = param.Integer(default=100, doc="""
        Width of the color swatches.""")

    value = param.Parameter(default=None, doc="The selected colormap.")

    value_name = param.String(default=None, doc="Name of the selected colormap.")

    _rename = {'options': 'items', 'value_name': None}

    _widget_type: ClassVar[type[Model]] = PaletteSelect

    @param.depends('value_name', watch=True, on_init=True)
    def _sync_value_name(self):
        if self.value_name and self.value_name in self.options:
            self.value = self.options[self.value_name]

    @param.depends('value', watch=True, on_init=True)
    def _sync_value(self):
        if self.value:
            idx = indexOf(self.value, self.values)
            self.value_name = self.labels[idx]

    def _process_param_change(self, params):
        if 'options' in params:
            options = []
            for name, cmap in params.pop('options').items():
                if 'matplotlib' in getattr(cmap, '__module__', ''):
                    N = getattr(cmap, 'N', 10)
                    samples = np.linspace(0, 1, N)
                    rgba_tmpl = 'rgba({0}, {1}, {2}, {3:.3g})'
                    cmap = [
                        rgba_tmpl.format(*(rgba[:3]*255).astype(int), rgba[-1])
                        for rgba in cmap(samples)
                    ]
                options.append((name, cmap))
            params['options'] = options
        if 'value' in params and not isinstance(params['value'], (str, type(None))):
            idx = indexOf(params['value'], self.values)
            params['value'] = self.labels[idx]
        return {
            self._property_mapping.get(p, p): v for p, v in params.items()
            if self._property_mapping.get(p, False) is not None
        }


class _MultiSelectBase(SingleSelectBase):

    value = param.List(default=[])

    width = param.Integer(default=300, allow_None=True, doc="""
      Width of this component. If sizing_mode is set to stretch
      or scale mode this will merely be used as a suggestion.""")

    description = param.String(default=None, doc="""
        An HTML string describing the function of this component.""")

    _supports_embed: ClassVar[bool] = False

    __abstract = True

    def _process_param_change(self, msg):
        msg = super(SingleSelectBase, self)._process_param_change(msg)
        labels, values = self.labels, self.values
        if 'value' in msg:
            msg['value'] = [labels[indexOf(v, values)] for v in msg['value']
                            if isIn(v, values)]

        if 'options' in msg:
            msg['options'] = labels
            if any(not isIn(v, values) for v in self.value):
                self.value = [v for v in self.value if isIn(v, values)]
        return msg

    def _process_property_change(self, msg):
        msg = super(SingleSelectBase, self)._process_property_change(msg)
        if 'value' in msg:
            labels = self.labels
            msg['value'] = [self._items[v] for v in msg['value']
                            if v in labels]
        msg.pop('options', None)
        return msg


class MultiSelect(_MultiSelectBase):
    """
    The `MultiSelect` widget allows selecting multiple values from a list of
    `options`.

    It falls into the broad category of multi-value, option-selection widgets
    that provide a compatible API and include the`CrossSelector`,
    `CheckBoxGroup` and `CheckButtonGroup` widgets.

    Reference: https://panel.holoviz.org/reference/widgets/MultiSelect.html

    :Example:

    >>> MultiSelect(
    ...     name='Frameworks', value=['Bokeh', 'Panel'],
    ...     options=['Bokeh', 'Dash', 'Panel', 'Streamlit', 'Voila'], size=8
    ... )
    """

    size = param.Integer(default=4, doc="""
        The number of items displayed at once (i.e. determines the
        widget height).""")

    _stylesheets: ClassVar[list[str]] = [f'{CDN_DIST}css/select.css']

    _widget_type: ClassVar[type[Model]] = _BkMultiSelect

    def __init__(self, **params):
        click_handler = params.pop('on_double_click', None)
        super().__init__(**params)
        self._dbl__click_handlers = [click_handler] if click_handler else []

    def _get_model(
        self, doc: Document, root: Optional[Model] = None,
        parent: Optional[Model] = None, comm: Optional[Comm] = None
    ) -> Model:
        model = super()._get_model(doc, root, parent, comm)
        self._register_events('dblclick_event', model=model, doc=doc, comm=comm)
        return model

    def _process_event(self, event: DoubleClickEvent) -> None:
        if event.option in self.labels:
            event.option = self._items[event.option]
            for handler in self._dbl__click_handlers:
                state.execute(partial(handler, event))

    def on_double_click(
        self, callback: Callable[[param.parameterized.Event], None | Awaitable[None]]
    ) -> param.parameterized.Watcher:
        """
        Register a callback to be executed when a `MultiSelect` option is double-clicked.

        The callback is given an `DoubleClickEvent` argument

        Example
        -------

        >>> select = pn.widgets.MultiSelect(options=["A", "B", "C"])
        >>> def handle_click(event):
        ...    print(f"Option {event.option} was double clicked.")
        >>> select.on_double_click(handle_click)

        Arguments
        ---------
        callback:
            The function to run on click events. Must accept a positional `Event` argument. Can
            be a sync or async function
        """
        self._dbl__click_handlers.append(callback)



class MultiChoice(_MultiSelectBase):
    """
    The `MultiChoice` widget allows selecting multiple values from a list of
    `options`.

    It falls into the broad category of multi-value, option-selection widgets
    that provide a compatible API and include the `MultiSelect`,
    `CrossSelector`, `CheckBoxGroup` and `CheckButtonGroup` widgets.

    The `MultiChoice` widget provides a much more compact UI than
    `MultiSelect`.

    Reference: https://panel.holoviz.org/reference/widgets/MultiChoice.html

    :Example:

    >>> MultiChoice(
    ...     name='Favourites', value=['Panel', 'hvPlot'],
    ...     options=['Panel', 'hvPlot', 'HoloViews', 'GeoViews', 'Datashader', 'Param', 'Colorcet'],
    ...     max_items=2
    ... )
    """

    delete_button = param.Boolean(default=True, doc="""
        Whether to display a button to delete a selected option.""")

    max_items = param.Integer(default=None, bounds=(1, None), doc="""
        Maximum number of options that can be selected.""")

    option_limit = param.Integer(default=None, bounds=(1, None), doc="""
        Maximum number of options to display at once.""")

    search_option_limit = param.Integer(default=None, bounds=(1, None), doc="""
        Maximum number of options to display at once if search string is entered.""")

    placeholder = param.String(default='', doc="""
        String displayed when no selection has been made.""")

    solid = param.Boolean(default=True, doc="""
        Whether to display widget with solid or light style.""")

    width = param.Integer(default=300, allow_None=True, doc="""
      Width of this component. If sizing_mode is set to stretch
      or scale mode this will merely be used as a suggestion.""")

    _widget_type: ClassVar[type[Model]] = _BkMultiChoice


class AutocompleteInput(SingleSelectBase):
    """
    The `AutocompleteInput` widget allows selecting multiple values from a list of
    `options`.

    It falls into the broad category of multi-value, option-selection widgets
    that provide a compatible API and include the `MultiSelect`,
    `CrossSelector`, `CheckBoxGroup` and `CheckButtonGroup` widgets.

    The `MultiChoice` widget provides a much more compact UI than
    `MultiSelect`.

    Reference: https://panel.holoviz.org/reference/widgets/AutocompleteInput.html

    :Example:

    >>> AutocompleteInput(
    ...     name='Study', options=['Biology', 'Chemistry', 'Physics'],
    ...     placeholder='Write your study here ...'
    ... )
    """

    case_sensitive = param.Boolean(default=True, doc="""
        Enable or disable case sensitivity.""")

    min_characters = param.Integer(default=2, doc="""
        The number of characters a user must type before
        completions are presented.""")

    placeholder = param.String(default='', doc="""
        Placeholder for empty input field.""")

    restrict = param.Boolean(default=True, doc="""
        Set to False in order to allow users to enter text that is not
        present in the list of completion strings.""")

    search_strategy = param.Selector(default='starts_with',
        objects=['starts_with', 'includes'], doc="""
        Define how to search the list of completion strings. The default option
        `"starts_with"` means that the user's text must match the start of a
        completion string. Using `"includes"` means that the user's text can
        match any substring of a completion string.""")

    value = param.Parameter(default='', allow_None=True, doc="""
      Initial or entered text value updated when <enter> key is pressed.""")

    value_input = param.String(default='', allow_None=True, doc="""
      Initial or entered text value updated on every key press.""")

    width = param.Integer(default=300, allow_None=True, doc="""
      Width of this component. If sizing_mode is set to stretch
      or scale mode this will merely be used as a suggestion.""")

    description = param.String(default=None, doc="""
        An HTML string describing the function of this component.""")

    _allows_values: ClassVar[bool] = False

    _allows_none: ClassVar[bool] = True

    _rename: ClassVar[Mapping[str, str | None]] = {'name': 'title', 'options': 'completions'}

    _widget_type: ClassVar[type[Model]] = _BkAutocompleteInput

    def _process_property_change(self, msg):
        if not self.restrict and 'value' in msg:
            try:
                return super()._process_property_change(msg)
            except Exception:
                return Widget._process_property_change(self, msg)
        return super()._process_property_change(msg)

    def _process_param_change(self, msg):
        if 'value' in msg and not self.restrict and not isIn(msg['value'], self.values):
            with param.parameterized.discard_events(self):
                props = super()._process_param_change(msg)
                self.value = props['value'] = msg['value']
        else:
            props = super()._process_param_change(msg)
        return props


class _RadioGroupBase(SingleSelectBase):

    _supports_embed = False

    _rename: ClassVar[Mapping[str, str | None]] = {
        'name': None, 'options': 'labels', 'value': 'active'
    }

    _source_transforms = {'value': "source.labels[value]"}

    _target_transforms = {'value': "target.labels.indexOf(value)"}

    __abstract = True

    def _process_param_change(self, msg):
        msg = super(SingleSelectBase, self)._process_param_change(msg)
        values = self.values
        if 'active' in msg:
            value = msg['active']
            if value in values:
                msg['active'] = indexOf(value, values)
            else:
                if self.value is not None:
                    self.value = None
                msg['active'] = None

        if 'labels' in msg:
            msg['labels'] = self.labels
            value = self.value
            if not isIn(value, values):
                self.value = None
        return msg

    def _process_property_change(self, msg):
        msg = super(SingleSelectBase, self)._process_property_change(msg)
        if 'value' in msg:
            index = msg['value']
            if index is None:
                msg['value'] = None
            else:
                msg['value'] = list(self.values)[index]
        return msg

    def _get_embed_state(self, root, values=None, max_opts=3):
        if values is None:
            values = self.values
        elif any(v not in self.values for v in values):
            raise ValueError("Supplied embed states were not found in "
                             f"the {type(self).__name__} widgets values list.")
        return (self, self._models[root.ref['id']][0], values,
                lambda x: x.active, 'active', 'cb_obj.active')



class RadioButtonGroup(_RadioGroupBase, _ButtonBase, TooltipMixin):
    """
    The `RadioButtonGroup` widget allows selecting from a list or dictionary
    of values using a set of toggle buttons.

    It falls into the broad category of single-value, option-selection widgets
    that provide a compatible API and include the `RadioBoxGroup`, `Select`,
    and `DiscreteSlider` widgets.

    Reference: https://panel.holoviz.org/reference/widgets/RadioButtonGroup.html

    :Example:

    >>> RadioButtonGroup(
    ...     name='Plotting library', options=['Matplotlib', 'Bokeh', 'Plotly'],
    ...     button_type='success'
    ... )
    """

    orientation = param.Selector(default='horizontal',
        objects=['horizontal', 'vertical'], doc="""
        Button group orientation, either 'horizontal' (default) or 'vertical'.""")

    _rename: ClassVar[Mapping[str, str | None]] = {**_RadioGroupBase._rename, **TooltipMixin._rename}

    _source_transforms = {
        'value': "source.labels[value]", 'button_style': None, 'description': None
    }

    _supports_embed: ClassVar[bool] = True

    _widget_type: ClassVar[type[Model]] = _BkRadioButtonGroup



class RadioBoxGroup(_RadioGroupBase):
    """
    The `RadioBoxGroup` widget allows selecting from a list or dictionary of
    values using a set of checkboxes.

    It falls into the broad category of single-value, option-selection widgets
    that provide a compatible API and include the `RadioButtonGroup`, `Select`
    and `DiscreteSlider` widgets.

    Reference: https://panel.holoviz.org/reference/widgets/RadioBoxGroup.html

    :Example:

    >>> RadioBoxGroup(
    ...     name='Sponsor', options=['Anaconda', 'Blackstone'], inline=True
    ... )
    """

    inline = param.Boolean(default=False, doc="""
        Whether the items be arrange vertically (``False``) or
        horizontally in-line (``True``).""")

    _supports_embed: ClassVar[bool] = True

    _widget_type: ClassVar[type[Model]] = _BkRadioBoxGroup



class _CheckGroupBase(SingleSelectBase):

    value = param.List(default=[])

    _rename: ClassVar[Mapping[str, str | None]] = {'name': None, 'options': 'labels', 'value': 'active'}

    _source_transforms = {'value': "value.map((index) => source.labels[index])"}

    _target_transforms = {'value': "value.map((label) => target.labels.indexOf(label))"}

    _supports_embed = False

    __abstract = True

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        values = self.values
        if 'active' in msg:
            msg['active'] = [indexOf(v, values) for v in msg['active']
                             if isIn(v, values)]
        if 'labels' in msg:
            msg['labels'] = self.labels
            if any(not isIn(v, values) for v in self.value):
                self.value = [v for v in self.value if isIn(v, values)]
            msg["active"] = [indexOf(v, values) for v in self.value
                             if isIn(v, values)]
        msg.pop('title', None)
        return msg

    def _process_property_change(self, msg):
        msg = super(SingleSelectBase, self)._process_property_change(msg)
        if 'value' in msg:
            values = self.values
            msg['value'] = [values[a] for a in msg['value']]
        return msg



class CheckButtonGroup(_CheckGroupBase, _ButtonBase, TooltipMixin):
    """
    The `CheckButtonGroup` widget allows selecting between a list of options
    by toggling the corresponding buttons.

    It falls into the broad category of multi-option selection widgets that
    provide a compatible API and include the `MultiSelect`, `CrossSelector`
    and `CheckBoxGroup` widgets.

    Reference: https://panel.holoviz.org/reference/widgets/CheckButtonGroup.html

    :Example:

    >>> CheckButtonGroup(
    ...     name='Regression Models', value=['Lasso', 'Ridge'],
    ...     options=['Lasso', 'Linear', 'Ridge', 'Polynomial']
    ... )
    """

    orientation = param.Selector(default='horizontal',
        objects=['horizontal', 'vertical'], doc="""
        Button group orientation, either 'horizontal' (default) or 'vertical'.""")

    _rename: ClassVar[Mapping[str, str | None]] = {**_CheckGroupBase._rename, **TooltipMixin._rename}

    _source_transforms = {
        'value': "value.map((index) => source.labels[index])", 'button_style': None,
        'description': None
    }

    _widget_type: ClassVar[type[Model]] = _BkCheckboxButtonGroup


class CheckBoxGroup(_CheckGroupBase):
    """
    The `CheckBoxGroup` widget allows selecting between a list of options by
    ticking the corresponding checkboxes.

    It falls into the broad category of multi-option selection widgets that
    provide a compatible API and include the `MultiSelect`, `CrossSelector`
    and `CheckButtonGroup` widgets.

    Reference: https://panel.holoviz.org/reference/widgets/CheckBoxGroup.html

    :Example:

    >>> CheckBoxGroup(
    ...     name='Fruits', value=['Apple', 'Pear'], options=['Apple', 'Banana', 'Pear', 'Strawberry'],
    ...     inline=True
    ... )
    """

    inline = param.Boolean(default=False, doc="""
        Whether the items be arrange vertically (``False``) or
        horizontally in-line (``True``).""")

    _widget_type: ClassVar[type[Model]] = _BkCheckboxGroup



class ToggleGroup(SingleSelectBase):
    """This class is a factory of ToggleGroup widgets.

    A ToggleGroup is a group of widgets which can be switched 'on' or 'off'.

    Two types of widgets are available through the widget_type argument :
        * `'button'` (default)
        * `'box'`

    Two different behaviors are available through behavior argument:
        * 'check' (default) : boolean
           Any number of widgets can be selected. In this case value
           is a 'list' of objects.
        * 'radio' : boolean
           One and only one widget is switched on. In this case value
           is an 'object'.
    """

    _widgets_type = ['button', 'box']
    _behaviors = ['check', 'radio']

    def __new__(cls, widget_type='button', behavior='check', **params):
        if widget_type not in ToggleGroup._widgets_type:
            raise ValueError(f'widget_type {widget_type} is not valid. Valid options are {ToggleGroup._widgets_type}')
        if behavior not in ToggleGroup._behaviors:
            raise ValueError(f'behavior {widget_type} is not valid. Valid options are {ToggleGroup._behaviors}')

        if behavior == 'check':
            if widget_type == 'button':
                return CheckButtonGroup(**params)
            else:
                return CheckBoxGroup(**params)
        else:
            if isinstance(params.get('value'), list):
                raise ValueError('Radio buttons require a single value, '
                                 'found: {}'.format(params['value']))
            if widget_type == 'button':
                return RadioButtonGroup(**params)
            else:
                return RadioBoxGroup(**params)



class CrossSelector(CompositeWidget, MultiSelect):
    """
    A composite widget which allows selecting from a list of items
    by moving them between two lists. Supports filtering values by
    name to select them in bulk.

    Reference: https://panel.holoviz.org/reference/widgets/CrossSelector.html

    :Example:

    >>> CrossSelector(
    ...     name='Fruits', value=['Apple', 'Pear'],
    ...     options=['Apple', 'Banana', 'Pear', 'Strawberry']
    ... )
    """

    width = param.Integer(default=600, allow_None=True, doc="""
        The number of options shown at once (note this is the
        only way to control the height of this widget)""")

    height = param.Integer(default=200, allow_None=True, doc="""
        The number of options shown at once (note this is the
        only way to control the height of this widget)""")

    filter_fn = param.Callable(default=re.search, doc="""
        The filter function applied when querying using the text
        fields, defaults to re.search. Function is two arguments, the
        query or pattern and the item label.""")

    size = param.Integer(default=10, doc="""
        The number of options shown at once (note this is the only way
        to control the height of this widget)""")

    definition_order = param.Integer(default=True, doc="""
       Whether to preserve definition order after filtering. Disable
       to allow the order of selection to define the order of the
       selected list.""")

    def __init__(self, **params):
        super().__init__(**params)
        # Compute selected and unselected values

        labels, values = self.labels, self.values
        selected = [
            labels[indexOf(v, values)] for v in params.get('value', [])
            if isIn(v, values)
        ]
        unselected = [k for k in labels if k not in selected]
        layout = dict(
            sizing_mode='stretch_both', margin=0
        )
        self._lists = {
            False: MultiSelect(options=unselected, size=self.size, **layout),
            True: MultiSelect(options=selected, size=self.size, **layout)
        }
        self._lists[False].param.watch(self._update_selection, 'value')
        self._lists[True].param.watch(self._update_selection, 'value')

        # Define buttons
        self._buttons = {
            False: Button(name='\u276e\u276e', width=50),
            True: Button(name='\u276f\u276f', width=50)
        }

        self._buttons[False].param.watch(self._apply_selection, 'clicks')
        self._buttons[True].param.watch(self._apply_selection, 'clicks')

        # Define search
        self._search = {
            False: TextInput(
                placeholder='Filter available options',
                margin=(0, 0, 10, 0), width_policy='max'
            ),
            True: TextInput(
                placeholder='Filter selected options',
                margin=(0, 0, 10, 0), width_policy='max'
            )
        }
        self._search[False].param.watch(self._filter_options, 'value_input')
        self._search[True].param.watch(self._filter_options, 'value_input')

        self._placeholder = TextAreaInput(
            placeholder=("To select an item highlight it on the left "
                         "and use the arrow button to move it to the right."),
            disabled=True, **layout
        )
        right = self._lists[True] if self.value else self._placeholder

        # Define Layout
        self._unselected = Column(self._search[False], self._lists[False], **layout)
        self._selected = Column(self._search[True], right, **layout)
        buttons = Column(self._buttons[True], self._buttons[False], margin=(0, 5), align='center')

        self._composite[:] = [
            self._unselected, buttons, self._selected
        ]

        self._selections = {False: [], True: []}
        self._query = {False: '', True: ''}

        self._update_disabled()
        self._update_width()

    @param.depends('width', watch=True)
    def _update_width(self):
        width = int(self.width // 2. - 50)
        self._search[False].width = width
        self._search[True].width = width
        self._lists[False].width = width
        self._lists[True].width = width

    @param.depends('size', watch=True)
    def _update_size(self):
        self._lists[False].size = self.size
        self._lists[True].size = self.size

    @param.depends('disabled', watch=True)
    def _update_disabled(self):
        self._buttons[False].disabled = self.disabled
        self._buttons[True].disabled = self.disabled

    @param.depends('value', watch=True)
    def _update_value(self):
        labels, values = self.labels, self.values
        selected = [labels[indexOf(v, values)] for v in self.value
                    if isIn(v, values)]
        unselected = [k for k in labels if k not in selected]
        self._lists[True].options = selected
        self._lists[True].value = []
        self._lists[False].options = unselected
        self._lists[False].value = []
        if len(self._lists[True].options) and self._selected[-1] is not self._lists[True]:
            self._selected[-1] = self._lists[True]
        elif not len(self._lists[True].options) and self._selected[-1] is not self._placeholder:
            self._selected[-1] = self._placeholder

    @param.depends('options', watch=True)
    def _update_options(self):
        """
        Updates the options of each of the sublists after the options
        for the whole widget are updated.
        """
        self._selections[False] = []
        self._selections[True] = []
        self._update_value()

    def _apply_filters(self):
        self._apply_query(False)
        self._apply_query(True)

    def _filter_options(self, event):
        """
        Filters unselected options based on a text query event.
        """
        selected = event.obj is self._search[True]
        self._query[selected] = event.new
        self._apply_query(selected)

    def _apply_query(self, selected):
        query = self._query[selected]
        other = self._lists[not selected].labels
        labels = self.labels
        if self.definition_order:
            options = [k for k in labels if k not in other]
        else:
            options = self._lists[selected].values
        if not query:
            self._lists[selected].options = options
            self._lists[selected].value = []
        else:
            try:
                matches = [o for o in options if self.filter_fn(query, o)]
            except Exception:
                matches = []
            self._lists[selected].options = options if options else []
            self._lists[selected].value = [m for m in matches]

    def _update_selection(self, event):
        """
        Updates the current selection in each list.
        """
        selected = event.obj is self._lists[True]
        self._selections[selected] = [v for v in event.new if v != '']

    def _apply_selection(self, event):
        """
        Applies the current selection depending on which button was
        pressed.
        """
        selected = event.obj is self._buttons[True]

        new = {k: self._items[k] for k in self._selections[not selected]}
        old = self._lists[selected].options
        other = self._lists[not selected].options

        merged = {k: k for k in list(old)+list(new)}
        leftovers = {k: k for k in other if k not in new}
        self._lists[selected].options = merged if merged else {}
        self._lists[not selected].options = leftovers if leftovers else {}
        if len(self._lists[True].options):
            self._selected[-1] = self._lists[True]
        else:
            self._selected[-1] = self._placeholder
        self.value = [self._items[o] for o in self._lists[True].options if o != '']
        self._apply_filters()

    def _get_model(self, doc, root=None, parent=None, comm=None):
        return self._composite._get_model(doc, root, parent, comm)



import param

import panel as pn

from panel.custom import ReactComponent


class MuiComponent(ReactComponent):

    _importmap = {
        "imports": {
            "@mui/material/": "https://esm.sh/@mui/material@5.15.16/",
        }
    }

class Button(MuiComponent):

    label = param.String()

    variant = param.Selector(default='contained', objects=['text', 'contained', 'outlined'])

    _esm = 'mui_button.js'

class DiscreteSlider(MuiComponent):

    marks = param.List(default=[])

    value = param.Number(default=20)

    _esm = 'mui_slider.js'


b = Button()
s = DiscreteSlider(marks=[
    {
        'value': 0,
        'label': '0°C',
    },
    {
        'value': 20,
        'label': '20°C',
    },
    {
        'value': 37,
        'label': '37°C',
    },
    {
        'value': 100,
        'label': '100°C',
    },
])

pn.Row(
    pn.Param(b.param, parameters=['label', 'variant']),
    pn.Column(b, s)
).servable()



import param
import pytest

import panel.custom  # To get the custom Viewable

from panel import config
from panel.interact import interactive
from panel.pane import Markdown, Str, panel
from panel.param import ParamMethod
from panel.viewable import (
    Child, Children, Viewable, Viewer, is_viewable_param,
)

from .util import jb_available

all_viewables = [w for w in param.concrete_descendents(Viewable).values()
               if not w.__name__.startswith('_') and
               not issubclass(w, interactive)]


class ExampleViewer(Viewer):
    value = param.String()

    def __panel__(self):
        return self.value


class ExampleViewerWithDeps(Viewer):
    value = param.String()

    @param.depends('value')
    def __panel__(self):
        return self.value


@jb_available
def test_viewable_ipywidget():
    pane = Str('A')
    with config.set(comms='ipywidgets'):
        data, metadata = pane._repr_mimebundle_()
    assert 'application/vnd.jupyter.widget-view+json' in data

@pytest.mark.parametrize('viewable', all_viewables)
def test_viewable_signature(viewable):
    from inspect import Parameter, signature
    parameters = signature(viewable).parameters
    if getattr(getattr(viewable, '_param__private', object), 'signature', None):
        pytest.skip('Signature already set by Param')
    assert 'params' in parameters
    try:
        assert parameters['params'] == Parameter('params', Parameter.VAR_KEYWORD, annotation='Any')
    except Exception:
        assert parameters['params'] == Parameter('params', Parameter.VAR_KEYWORD)

def test_Viewer_not_initialized():
    class Test(Viewer):
        def __panel__(self):
            return "# Test"

    test = panel(Test)
    assert test.object == "# Test"

    # Confirm that initialized also work
    test = panel(Test())
    assert test.object == "# Test"

def test_viewer_wraps_panel():
    tv = ExampleViewer(value="hello")

    view = tv._create_view()
    assert isinstance(view, Markdown)
    assert view.object == "hello"

def test_viewer_wraps_panel_with_deps(document, comm):
    tv = ExampleViewerWithDeps(value="hello")

    view = tv._create_view()

    view.get_root(document, comm)

    assert isinstance(view, ParamMethod)
    assert view._pane.object == "hello"

    tv.value = "goodbye"

    assert view._pane.object == "goodbye"

def test_viewer_with_deps_resolved_by_panel_func(document, comm):
    tv = ExampleViewerWithDeps(value="hello")

    view = panel(tv)

    view.get_root(document, comm)

    assert isinstance(view, ParamMethod)
    assert view._pane.object == "hello"

def test_non_viewer_class():
    # This test checks that a class with __panel__ (other than Viewer)
    # does not raise a TypeError: issubclass() arg 1 must be a class

    class Example:
        def __panel__(self):
            return 42

    panel(Example())

@pytest.mark.parametrize('viewable', all_viewables)
def test_clone(viewable):
    v = Viewable()
    clone = v.clone()

    assert ([(k, v) for k, v in sorted(v.param.values().items()) if k not in ('name')] ==
            [(k, v) for k, v in sorted(clone.param.values().items()) if k not in ('name')])

def test_clone_with_non_defaults():
    v= Viewable(loading=True)
    clone = v.clone()

    assert ([(k, v) for k, v in sorted(v.param.values().items()) if k not in ('name')] ==
            [(k, v) for k, v in sorted(clone.param.values().items()) if k not in ('name')])

def test_is_viewable_parameter():
    class Example(param.Parameterized):
        p_dict = param.Dict()
        p_child = Child()
        p_children = Children()

        # ClassSelector
        c_viewable = param.ClassSelector(class_=Viewable)
        c_viewables = param.ClassSelector(class_=(Viewable,))
        c_none = param.ClassSelector(class_=None)
        c_tuple = param.ClassSelector(class_=tuple)
        c_list_tuple = param.ClassSelector(class_=(list, tuple))

        # List
        l_no_item_type = param.List()
        l_item_type_viewable = param.List(item_type=Viewable)
        l_item_type_not_viewable = param.List(item_type=tuple)

        l_item_types_viewable = param.List(item_type=(Viewable,))
        l_item_types_not_viewable = param.List(item_type=(tuple,))
        l_item_types_not_viewable2 = param.List(item_type=(list, tuple,))

    example = Example()

    assert not is_viewable_param(example.param.p_dict)
    assert is_viewable_param(example.param.p_child)
    assert is_viewable_param(example.param.p_children)

    # ClassSelector
    assert is_viewable_param(example.param.c_viewable)
    assert is_viewable_param(example.param.c_viewables)
    assert not is_viewable_param(example.param.c_none)
    assert not is_viewable_param(example.param.c_tuple)
    assert not is_viewable_param(example.param.c_list_tuple)

    # List
    assert not is_viewable_param(example.param.l_no_item_type)
    assert not is_viewable_param(example.param.l_no_item_type)
    assert is_viewable_param(example.param.l_item_type_viewable)
    assert not is_viewable_param(example.param.l_item_type_not_viewable)

    assert is_viewable_param(example.param.l_item_types_viewable)
    assert not is_viewable_param(example.param.l_item_types_not_viewable)
    assert not is_viewable_param(example.param.l_item_types_not_viewable2)



import param

import panel as pn

from panel.custom import JSComponent

pn.extension()

class JSSlideshow(JSComponent):

    index = param.Integer(default=0)

    _esm = """
	export function render({ model }) {
      const img = document.createElement('img')
      img.addEventListener('click', () => { model.index += 1 })
      const update = () => {
        img.src = `https://picsum.photos/800/300?image=${model.index}`
      }
      model.watch(update, 'index')
      update()
      return img
    }
	"""


class PySlideshow(JSComponent):

    index = param.Integer(default=0)

    _esm = """
	export function render({ model }) {
      const img = document.createElement('img')
      img.addEventListener('click', (event) => model.send_event('click', event))
      const update = () => {
        img.src = `https://picsum.photos/800/300?image=${model.index}`
      }
      model.watch(update, 'index')
      update()
      return img
    }
	"""

    def _handle_click(self, event):
        self.index += 1


pn.Row(
    JSSlideshow(),
    PySlideshow()
).servable()
